@@ -10,19 +10,20 @@
    GNU General Public License for more details.
 */
 
-/* Author's email: simon@thekelleys.org.uk */
-
-#define VERSION "2.25"
+#define VERSION "2.31"
 
 #define FTABSIZ 150 /* max number of outstanding requests */
 #define MAX_PROCS 20 /* max no children for TCP requests */
 #define CHILD_LIFETIME 150 /* secs 'till terminated (RFC1035 suggests > 120s) */
 #define EDNS_PKTSZ 1280 /* default max EDNS.0 UDP packet from RFC2671 */
 #define TIMEOUT 20 /* drop UDP queries after TIMEOUT seconds */
+#define LEASE_RETRY 60 /* on error, retry writing leasefile after LEASE_RETRY seconds */
 #define LOGRATE 120 /* log table overflows every LOGRATE seconds */
 #define CACHESIZ 150 /* default cache size */
-#define MAXTOK 50 /* token in DHCP leases */
 #define MAXLEASES 150 /* maximum number of DHCP leases */
+#define PING_WAIT 3 /* wait for ping address-in-use test */
+#define PING_CACHE_TIME 30 /* Ping test assumed to be valid this long. */
+#define DHCP_PACKET_MAX 16384 /* hard limit on DHCP packet size */
 #define SMALLDNAME 40 /* most domain names are smaller than this */
 #define HOSTSFILE "/etc/hosts"
 #define ETHERSFILE "/etc/ethers"
@@ -45,8 +46,6 @@
 #define DEFLEASE 3600 /* default lease time, 1 hour */
 #define CHUSER "nobody"
 #define CHGRP "dip"
-#define IP6INTERFACES "/proc/net/if_inet6"
-#define UPTIME "/proc/uptime"
 #define DHCP_SERVER_PORT 67
 #define DHCP_CLIENT_PORT 68
 
@@ -80,7 +79,8 @@
 
 /* Get linux C library versions. */
 #if defined(__linux__) && !defined(__UCLIBC__) && !defined(__uClinux__)
-#  include <libio.h> 
+/*#  include <libio.h> */
+#  include <features.h> 
 #endif
 
 
@@ -88,26 +88,25 @@
    new system, you may want to edit these. 
    May replace this with Autoconf one day. 
 
-HAVE_LINUX_IPV6_PROC
-   define this to do IPv6 interface discovery using
-   proc/net/if_inet6 ala LINUX. 
+HAVE_LINUX_NETWORK
+   define this to do networking the Linux way. When it's defined, the code will
+   use IP_PKTINFO, Linux capabilities and the RTnetlink system. If it's not defined,
+   a few facilities will be lost, namely support for multiple addresses on an interface,
+   DNS query retransmission, and (on some systems) wildcard interface binding.
 
 HAVE_BROKEN_RTC
-   define this on embeded systems which don't have an RTC
-   which keeps time over reboots. Causes dnsmasq to use uptime()
-   for timing, and keep relative time values in its leases file.
-   Also enables "Flash disk mode". Normally, dnsmasq tries very hard to 
-   keep the on-disk leases file up-to-date: rewriting it after every change.
-   When HAVE_BROKEN_RTC is in effect, a different regime is used:
-   The leases file is written when dnsmasq terminates, when it receives
-   SIGALRM, when a brand new lease is allocated, or every n seconds, 
-   where n is one third  of the smallest time configured for leases 
-   in a --dhcp-range or --dhcp-host option.
+   define this on embedded systems which don't have an RTC
+   which keeps time over reboots. Causes dnsmasq to use uptime
+   for timing, and keep lease lengths rather than expiry times
+   in its leases file. This also make dnsmasq "flash disk friendly".
+   Normally, dnsmasq tries very hard to keep the on-disk leases file
+   up-to-date: rewriting it after every renewal.  When HAVE_BROKEN_RTC 
+   is in effect, the lease file is only written when a new lease is 
+   created, or an old one destroyed. (Because those are the only times 
+   it changes.) This vastly reduces the number of file writes, and makes
+   it viable to keep the lease file on a flash filesystem.
    NOTE: when enabling or disabling this, be sure to delete any old
    leases file, otherwise dnsmasq may get very confused.
-   This configuration currently only works on Linux, but could be made to
-   work on other systems by teaching dnsmasq_time() in utils.c how to
-   read the system uptime.
 
 HAVE_ISC_READER 
    define this to include the old ISC dhcpcd integration. Note that you cannot
@@ -136,18 +135,6 @@ HAVE_DEV_URANDOM
 HAVE_SOCKADDR_SA_LEN
    define this if struct sockaddr has sa_len field (*BSD) 
 
-HAVE_PSELECT
-   If your C library implements pselect, define this.
-
-HAVE_BPF
-   If your OS implements Berkeley Packet filter, define this.
-
-HAVE_RTNETLINK
-   If your OS has the Linux Routing netlink socket API and suitable
-   C library headers, define this. Note that the code will fall
-   back to the Berkley API at runtime if netlink support is not
-   configured into the kernel.
-
 HAVE_DBUS
    Define this if you want to link against libdbus, and have dnsmasq
    define some methods to allow (re)configuration of the upstream DNS 
@@ -155,12 +142,11 @@ HAVE_DBUS
 
 NOTES:
    For Linux you should define 
-      HAVE_LINUX_IPV6_PROC 
+      HAVE_LINUX_NETWORK
       HAVE_GETOPT_LONG
       HAVE_RANDOM
       HAVE_DEV_RANDOM
       HAVE_DEV_URANDOM
-      HAVE_RTNETLINK
   you should NOT define 
       HAVE_ARC4RANDOM
       HAVE_SOCKADDR_SA_LEN
@@ -168,10 +154,8 @@ NOTES:
    For *BSD systems you should define 
      HAVE_SOCKADDR_SA_LEN
      HAVE_RANDOM
-     HAVE_BPF
    you should NOT define  
-     HAVE_LINUX_IPV6_PROC 
-     HAVE_RTNETLINK
+     HAVE_LINUX_NETWORK 
    and you MAY define  
      HAVE_ARC4RANDOM - OpenBSD and FreeBSD and NetBSD version 2.0 or later
      HAVE_DEV_URANDOM - OpenBSD and FreeBSD and NetBSD
@@ -182,10 +166,10 @@ NOTES:
 
 */
 
-/* platform independent options. */
-#undef HAVE_BROKEN_RTC
-#define HAVE_ISC_READER
-#undef HAVE_DBUS
+/* platform independent options- uncomment to enable */
+/* #define HAVE_BROKEN_RTC */
+/* #define HAVE_ISC_READER */
+/* #define HAVE_DBUS */
 
 #if defined(HAVE_BROKEN_RTC) && defined(HAVE_ISC_READER)
 #  error HAVE_ISC_READER is not compatible with HAVE_BROKEN_RTC
@@ -195,76 +179,49 @@ NOTES:
 
 /* Must preceed __linux__ since uClinux defines __linux__ too. */
 #if defined(__uClinux__)
-#define HAVE_LINUX_IPV6_PROC
+#define HAVE_LINUX_NETWORK
 #define HAVE_GETOPT_LONG
-#define HAVE_RTNETLINK
 #undef HAVE_ARC4RANDOM
 #define HAVE_RANDOM
 #define HAVE_DEV_URANDOM
 #define HAVE_DEV_RANDOM
 #undef HAVE_SOCKADDR_SA_LEN
-#undef HAVE_PSELECT
 /* Never use fork() on uClinux. Note that this is subtly different from the
    --keep-in-foreground option, since it also  suppresses forking new 
    processes for TCP connections. It's intended for use on MMU-less kernels. */
 #define NO_FORK
 
 #elif defined(__UCLIBC__)
-#define HAVE_LINUX_IPV6_PROC
+#define HAVE_LINUX_NETWORK
 #if defined(__UCLIBC_HAS_GNU_GETOPT__) || \
    ((__UCLIBC_MAJOR__==0) && (__UCLIBC_MINOR__==9) && (__UCLIBC_SUBLEVEL__<21))
 #    define HAVE_GETOPT_LONG
-#  else
+#else
 #    undef HAVE_GETOPT_LONG
-#  endif
-#define HAVE_RTNETLINK
+#endif
 #undef HAVE_ARC4RANDOM
 #define HAVE_RANDOM
 #define HAVE_DEV_URANDOM
 #define HAVE_DEV_RANDOM
 #undef HAVE_SOCKADDR_SA_LEN
-#undef HAVE_PSELECT
-#if !defined(__ARCH_HAS_MMU__)
+#if !defined(__ARCH_HAS_MMU__) && !defined(__UCLIBC_HAS_MMU__)
 #  define NO_FORK
 #endif
-#if !defined(__UCLIBC_HAS_IPV6__)
-#  define NO_IPV6
+#if defined(__UCLIBC_HAS_IPV6__)
+#  ifndef IPV6_V6ONLY
+#    define IPV6_V6ONLY 26
+#  endif
 #endif
 
-/* libc5 - must precede __linux__ too */
-/* Note to build a libc5 binary on a modern Debian system:
-   install the packages altgcc libc5 and libc5-altdev 
-   then run "make CC=i486-linuxlibc1-gcc" */
-/* Note that compling dnsmasq 2.x under libc5 and kernel 2.0.x
-   is probably doomed - no packet socket for starters. */
-#elif defined(__linux__) && \
-      defined(_LINUX_C_LIB_VERSION_MAJOR) && \
-      (_LINUX_C_LIB_VERSION_MAJOR == 5 )
-#undef HAVE_IPV6
-#undef HAVE_LINUX_IPV6_PROC
-#undef HAVE_RTNETLINK
-#define HAVE_GETOPT_LONG
-#undef HAVE_ARC4RANDOM
-#define HAVE_RANDOM
-#define HAVE_DEV_URANDOM
-#define HAVE_DEV_RANDOM
-#undef HAVE_SOCKADDR_SA_LEN
-#undef HAVE_PSELECT
-/* Fix various misfeatures of libc5 headers */
-typedef unsigned long in_addr_t; 
-typedef size_t socklen_t;
-
 /* This is for glibc 2.x */
 #elif defined(__linux__)
-#define HAVE_LINUX_IPV6_PROC
-#define HAVE_RTNETLINK
+#define HAVE_LINUX_NETWORK
 #define HAVE_GETOPT_LONG
 #undef HAVE_ARC4RANDOM
 #define HAVE_RANDOM
 #define HAVE_DEV_URANDOM
 #define HAVE_DEV_RANDOM
 #undef HAVE_SOCKADDR_SA_LEN
-#define HAVE_PSELECT
 /* glibc < 2.2  has broken Sockaddr_in6 so we have to use our own. */
 /* glibc < 2.2 doesn't define in_addr_t */
 #if defined(__GLIBC__) && (__GLIBC__ == 2) && \
@@ -274,8 +231,7 @@ typedef unsigned long in_addr_t;
 #endif
 
 #elif defined(__FreeBSD__) || defined(__OpenBSD__)
-#undef HAVE_LINUX_IPV6_PROC
-#undef HAVE_RTNETLINK
+#undef HAVE_LINUX_NETWORK
 /* Later verions of FreeBSD have getopt_long() */
 #if defined(optional_argument) && defined(required_argument)
 #   define HAVE_GETOPT_LONG
@@ -286,51 +242,29 @@ typedef unsigned long in_addr_t;
 #define HAVE_RANDOM
 #define HAVE_DEV_URANDOM
 #define HAVE_SOCKADDR_SA_LEN
-#undef HAVE_PSELECT
-#define HAVE_BPF
 
 #elif defined(__APPLE__)
-#undef HAVE_LINUX_IPV6_PROC
-#undef HAVE_RTNETLINK
+#undef HAVE_LINUX_NETWORK
 #undef HAVE_GETOPT_LONG
 #define HAVE_ARC4RANDOM
 #define HAVE_RANDOM
 #define HAVE_DEV_URANDOM
 #define HAVE_SOCKADDR_SA_LEN
-#undef HAVE_PSELECT
-#define HAVE_BPF
 /* Define before sys/socket.h is included so we get socklen_t */
 #define _BSD_SOCKLEN_T_
 /* This is not defined in Mac OS X arpa/nameserv.h */
 #define IN6ADDRSZ 16
  
 #elif defined(__NetBSD__)
-#undef HAVE_LINUX_IPV6_PROC
-#undef HAVE_RTNETLINK
+#undef HAVE_LINUX_NETWORK
 #define HAVE_GETOPT_LONG
 #undef HAVE_ARC4RANDOM
 #define HAVE_RANDOM
 #define HAVE_DEV_URANDOM
 #define HAVE_DEV_RANDOM
 #define HAVE_SOCKADDR_SA_LEN
-#undef HAVE_PSELECT
-#define HAVE_BPF
  
-/* env "LIBS=-lsocket -lnsl" make */
-#elif defined(__sun) || defined(__sun__)
-#undef HAVE_LINUX_IPV6_PROC
-#undef HAVE_RTNETLINK
-#undef HAVE_GETOPT_LONG
-#undef HAVE_ARC4RANDOM
-#define HAVE_RANDOM
-#undef HAVE_DEV_URANDOM
-#undef HAVE_DEV_RANDOM
-#undef HAVE_SOCKADDR_SA_LEN
-#undef HAVE_PSELECT
-#define HAVE_BPF
 #endif
-
-
 /* Decide if we're going to support IPv6 */
 /* We assume that systems which don't have IPv6
    headers don't have ntop and pton either */
@@ -351,5 +285,3 @@ typedef unsigned long in_addr_t;
 #  define ADDRSTRLEN 16 /* 4*3 + 3 dots + NULL */
 #endif
 
-
-
