@@ -1,4 +1,4 @@
-/* dnsmasq is Copyright (c) 2000-2005 Simon Kelley
+/* dnsmasq is Copyright (c) 2000-2006 Simon Kelley
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -10,36 +10,42 @@
    GNU General Public License for more details.
 */
 
-/* Author's email: simon@thekelleys.org.uk */
-
 #include "dnsmasq.h"
 
+struct iface_param {
+  struct in_addr relay, primary;
+  struct dhcp_context *current;
+  int ind;
+};
+
+static int complete_context(struct daemon *daemon, struct in_addr local, int if_index, 
+			    struct in_addr netmask, struct in_addr broadcast, void *vparam);
+
 void dhcp_init(struct daemon *daemon)
 {
   int fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
   struct sockaddr_in saddr;
-  int flags, oneopt = 1, zeroopt = 0;
+  int oneopt = 1;
   struct dhcp_config *configs, *cp;
 
   if (fd == -1)
-    die ("cannot create DHCP socket : %s", NULL);
+    die (_("cannot create DHCP socket : %s"), NULL);
   
-  if ((flags = fcntl(fd, F_GETFL, 0)) == -1 ||
-      fcntl(fd, F_SETFL, flags | O_NONBLOCK) == -1 ||
-#if defined(IP_PKTINFO)
+  if (!fix_fd(fd) ||
+#if defined(HAVE_LINUX_NETWORK)
       setsockopt(fd, SOL_IP, IP_PKTINFO, &oneopt, sizeof(oneopt)) == -1 ||
 #elif defined(IP_RECVIF)
       setsockopt(fd, IPPROTO_IP, IP_RECVIF, &oneopt, sizeof(oneopt)) == -1 ||
 #endif
       setsockopt(fd, SOL_SOCKET, SO_BROADCAST, &oneopt, sizeof(oneopt)) == -1)  
-    die("failed to set options on DHCP socket: %s", NULL);
+    die(_("failed to set options on DHCP socket: %s"), NULL);
   
   /* When bind-interfaces is set, there might be more than one dnmsasq
      instance binding port 67. That's Ok if they serve different networks.
      Need to set REUSEADDR to make this posible. */
   if ((daemon->options & OPT_NOWILD) &&
       setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &oneopt, sizeof(oneopt)) == -1)
-    die("failed to set SO_REUSEADDR on DHCP socket: %s", NULL);
+    die(_("failed to set SO_REUSEADDR on DHCP socket: %s"), NULL);
   
   saddr.sin_family = AF_INET;
   saddr.sin_port = htons(DHCP_SERVER_PORT);
@@ -49,45 +55,24 @@ void dhcp_init(struct daemon *daemon)
 #endif
 
   if (bind(fd, (struct sockaddr *)&saddr, sizeof(struct sockaddr_in)))
-    die("failed to bind DHCP server socket: %s", NULL);
+    die(_("failed to bind DHCP server socket: %s"), NULL);
 
   daemon->dhcpfd = fd;
 
-  if ((fd = socket (AF_INET, SOCK_RAW, IPPROTO_ICMP)) == -1 ||
-      (flags = fcntl(fd, F_GETFL, 0)) == -1 ||
-      fcntl(fd, F_SETFL, flags | O_NONBLOCK) == -1 ||
-      setsockopt(fd, SOL_SOCKET, SO_RCVBUF, &oneopt, sizeof(oneopt)) == -1 ||
-      setsockopt(fd, SOL_SOCKET, SO_DONTROUTE, &zeroopt, sizeof(zeroopt)) == -1)
-    die("cannot create ICMP raw socket: %s.", NULL);
-
-  daemon->dhcp_icmp_fd = fd;
-
-#ifdef HAVE_BPF
-  { 
-    int i = 0;
-    while (1) 
-      {
-	char filename[50];
-	sprintf(filename, "/dev/bpf%d", i++);
-	if ((fd = open(filename, O_RDWR, 0)) != -1)
-	  break;
-	if (errno != EBUSY)
-	  die("cannot create DHCP BPF socket: %s", NULL);
-      }	    
-  }
-#else
-  /* since we don't ever use the packet socket for reception,
-     and it receives copies of _all_ IP packets, then that data
-     will build up in kernel buffers, wasting memory. Set the
-     socket receive buffer size to one to avoid that. (zero is
-     rejected as non-sensical by some BSD kernels) */
-  if ((fd = socket(PF_PACKET, SOCK_DGRAM, htons(ETHERTYPE_IP))) == -1 ||
-      setsockopt(fd, SOL_SOCKET, SO_RCVBUF, &oneopt, sizeof(oneopt)) == -1)
-    die("cannot create DHCP packet socket: %s. "
-	"Is CONFIG_PACKET enabled in your kernel?", NULL);
-#endif
+#ifndef HAVE_LINUX_NETWORK
+  /* When we're not using capabilities, we need to do this here before
+     we drop root. Also, set buffer size small, to avoid wasting
+     kernel buffers */
   
-  daemon->dhcp_raw_fd = fd;
+  if (daemon->options & OPT_NO_PING)
+    daemon->dhcp_icmp_fd = -1;
+  else if ((daemon->dhcp_icmp_fd = make_icmp_sock()) == -1 ||
+	   setsockopt(daemon->dhcp_icmp_fd, SOL_SOCKET, SO_RCVBUF, &oneopt, sizeof(oneopt)) == -1 )
+    die(_("cannot create ICMP raw socket: %s."), NULL);
+  
+  /* Make BPF raw send socket */
+  init_bpf(daemon);
+#endif
   
   /* If the same IP appears in more than one host config, then DISCOVER
      for one of the hosts will get the address, but REQUEST will be NAKed,
@@ -95,75 +80,91 @@ void dhcp_init(struct daemon *daemon)
   for (configs = daemon->dhcp_conf; configs; configs = configs->next)
     for (cp = configs->next; cp; cp = cp->next)
       if ((configs->flags & cp->flags & CONFIG_ADDR) &&	configs->addr.s_addr == cp->addr.s_addr)
-	die("duplicate IP address %s in dhcp-config directive.", inet_ntoa(cp->addr));
+	die(_("duplicate IP address %s in dhcp-config directive."), inet_ntoa(cp->addr));
   
-  daemon->dhcp_packet = safe_malloc(sizeof(struct udp_dhcp_packet));
-  /* These two each hold a DHCP option max size 255
+  daemon->dhcp_packet.iov_len = sizeof(struct dhcp_packet); 
+  daemon->dhcp_packet.iov_base = safe_malloc(daemon->dhcp_packet.iov_len);
+    /* These two each hold a DHCP option max size 255
      and get a terminating zero added */
   daemon->dhcp_buff = safe_malloc(256);
   daemon->dhcp_buff2 = safe_malloc(256); 
   daemon->ping_results = NULL;
 }
-
+  
 void dhcp_packet(struct daemon *daemon, time_t now)
 {
-  struct udp_dhcp_packet *rawpacket = daemon->dhcp_packet;
-  struct dhcp_packet *mess = &rawpacket->data;
+  struct dhcp_packet *mess;
   struct dhcp_context *context;
   struct iname *tmp;
   struct ifreq ifr;
   struct msghdr msg;
-  struct iovec iov[2];
+  struct sockaddr_in dest;
   struct cmsghdr *cmptr;
-  int sz, newlen, iface_index = 0;
+  struct iovec iov;
+  ssize_t sz; 
+  int iface_index = 0, unicast_dest = 0;
   struct in_addr iface_addr;
-#ifdef HAVE_BPF
-  unsigned char iface_hwaddr[ETHER_ADDR_LEN];
-#endif
+  struct iface_param parm;
 
   union {
     struct cmsghdr align; /* this ensures alignment */
-#ifdef IP_PKTINFO
+#ifdef HAVE_LINUX_NETWORK
     char control[CMSG_SPACE(sizeof(struct in_pktinfo))];
 #else
     char control[CMSG_SPACE(sizeof(struct sockaddr_dl))];
 #endif
   } control_u;
   
-  iov[0].iov_base = (char *)mess;
-  iov[0].iov_len = sizeof(struct dhcp_packet);
-
   msg.msg_control = control_u.control;
   msg.msg_controllen = sizeof(control_u);
-  msg.msg_flags = 0;
   msg.msg_name = NULL;
   msg.msg_namelen = 0;
-  msg.msg_iov = iov;
+  msg.msg_iov = &daemon->dhcp_packet;
   msg.msg_iovlen = 1;
   
-  sz = recvmsg(daemon->dhcpfd, &msg, 0);
+  do
+    {
+      msg.msg_flags = 0;
+      while ((sz = recvmsg(daemon->dhcpfd, &msg, MSG_PEEK)) == -1 && errno == EINTR);
+    }
+  while (sz != -1 && (msg.msg_flags & MSG_TRUNC) &&
+	 expand_buf(&daemon->dhcp_packet, daemon->dhcp_packet.iov_len + 100));
   
-  if (sz < (int)(sizeof(*mess) - sizeof(mess->options)))
+  /* expand_buf may have moved buffer */
+  mess = daemon->dhcp_packet.iov_base;
+  msg.msg_controllen = sizeof(control_u);
+  msg.msg_flags = 0;
+  msg.msg_name = &dest;
+  msg.msg_namelen = sizeof(dest);
+
+  while ((sz = recvmsg(daemon->dhcpfd, &msg, 0)) && errno == EINTR);
+ 
+  if ((msg.msg_flags & MSG_TRUNC) ||
+      sz < (ssize_t)(sizeof(*mess) - sizeof(mess->options)))
     return;
   
-#if defined (IP_PKTINFO)
+#if defined (HAVE_LINUX_NETWORK)
   if (msg.msg_controllen < sizeof(struct cmsghdr))
     return;
   for (cmptr = CMSG_FIRSTHDR(&msg); cmptr; cmptr = CMSG_NXTHDR(&msg, cmptr))
     if (cmptr->cmsg_level == SOL_IP && cmptr->cmsg_type == IP_PKTINFO)
-      iface_index = ((struct in_pktinfo *)CMSG_DATA(cmptr))->ipi_ifindex;
-  
+      {
+	iface_index = ((struct in_pktinfo *)CMSG_DATA(cmptr))->ipi_ifindex;
+	if (((struct in_pktinfo *)CMSG_DATA(cmptr))->ipi_addr.s_addr != INADDR_BROADCAST)
+	  unicast_dest = 1;
+      }
+
   if (!(ifr.ifr_ifindex = iface_index) || 
       ioctl(daemon->dhcpfd, SIOCGIFNAME, &ifr) == -1)
     return;
-
+  
 #elif defined(IP_RECVIF)
   if (msg.msg_controllen < sizeof(struct cmsghdr))
     return;
   for (cmptr = CMSG_FIRSTHDR(&msg); cmptr; cmptr = CMSG_NXTHDR(&msg, cmptr))
     if (cmptr->cmsg_level == IPPROTO_IP && cmptr->cmsg_type == IP_RECVIF)
       iface_index = ((struct sockaddr_dl *)CMSG_DATA(cmptr))->sdl_index;
-  
+
   if (!iface_index || !if_indextoname(iface_index, ifr.ifr_name))
     return;
 
@@ -172,197 +173,133 @@ void dhcp_packet(struct daemon *daemon, time_t now)
     struct iname *name;
     for (name = daemon->if_names; name->isloop; name = name->next);
     strcpy(ifr.ifr_name, name->name);
+    iface_index = if_nametoindex(name->name);
   }
 #endif
 
-#ifdef HAVE_BPF
-  ifr.ifr_addr.sa_family = AF_LINK;
-  if (ioctl(daemon->dhcpfd, SIOCGIFADDR, &ifr) < 0)
-    return;
-  memcpy(iface_hwaddr, LLADDR((struct sockaddr_dl *)&ifr.ifr_addr), ETHER_ADDR_LEN);
-#endif
-  
   ifr.ifr_addr.sa_family = AF_INET;
   if (ioctl(daemon->dhcpfd, SIOCGIFADDR, &ifr) < 0 )
     return;
   iface_addr = ((struct sockaddr_in *) &ifr.ifr_addr)->sin_addr;
 
-  /* enforce available interface configuration */
-  for (tmp = daemon->if_except; tmp; tmp = tmp->next)
-    if (tmp->name && (strcmp(tmp->name, ifr.ifr_name) == 0))
-      return;
- 
   for (tmp = daemon->dhcp_except; tmp; tmp = tmp->next)
     if (tmp->name && (strcmp(tmp->name, ifr.ifr_name) == 0))
       return;
   
-  if (daemon->if_names || daemon->if_addrs)
-    {
-      for (tmp = daemon->if_names; tmp; tmp = tmp->next)
-	if (tmp->name && (strcmp(tmp->name, ifr.ifr_name) == 0))
-	  break;
-      if (!tmp)
-	for (tmp = daemon->if_addrs; tmp; tmp = tmp->next)
-	  if (tmp->addr.sa.sa_family == AF_INET && 
-	      tmp->addr.in.sin_addr.s_addr == iface_addr.s_addr)
-	    break;
-      if (!tmp)
-	return; 
-    }
+  if (!iface_check(daemon, AF_INET, (struct all_addr *)&iface_addr, ifr.ifr_name))
+    return;
   
   /* unlinked contexts are marked by context->current == context */
   for (context = daemon->dhcp; context; context = context->next)
     context->current = context;
   
-#ifdef HAVE_RTNETLINK
-  if (!netlink_process(daemon, iface_index, mess->giaddr, iface_addr, &context))
-#endif
-    {
-      struct in_addr iface_netmask, iface_broadcast;
-      
-      if (ioctl(daemon->dhcpfd, SIOCGIFNETMASK, &ifr) < 0)
-	return;
-      iface_netmask = ((struct sockaddr_in *) &ifr.ifr_addr)->sin_addr;
-      
-      if (ioctl(daemon->dhcpfd, SIOCGIFBRDADDR, &ifr) < 0)
-	return;
-      iface_broadcast = ((struct sockaddr_in *) &ifr.ifr_addr)->sin_addr;
-      
-      context = complete_context(daemon, iface_addr, NULL, iface_netmask, 
-				 iface_broadcast, mess->giaddr, iface_addr);
-    }
+  parm.relay = mess->giaddr;
+  parm.primary = iface_addr;
+  parm.current = NULL;
+  parm.ind = iface_index;
 
+  if (!iface_enumerate(daemon, &parm, complete_context, NULL))
+    return;
   lease_prune(NULL, now); /* lose any expired leases */
-  newlen = dhcp_reply(daemon, context, ifr.ifr_name, sz, now);
-  lease_update_file(0, now);
+  iov.iov_len = dhcp_reply(daemon, parm.current, ifr.ifr_name, (size_t)sz, now, unicast_dest);
+  lease_update_file(daemon, now);
   lease_update_dns(daemon);
+  lease_collect(daemon);
   
-  if (newlen == 0)
+  if (iov.iov_len == 0)
     return;
   
-  if (mess->giaddr.s_addr || mess->ciaddr.s_addr)
-    {
-      /* To send to BOOTP relay or configured client, use the IP packet */
-      
-      struct sockaddr_in dest;
-      dest.sin_family = AF_INET;
+  msg.msg_name = &dest;
+  msg.msg_namelen = sizeof(dest);
+  msg.msg_control = NULL;
+  msg.msg_controllen = 0;
+  msg.msg_iov = &iov;
+  iov.iov_base = daemon->dhcp_packet.iov_base;
+  
+  /* packet buffer may have moved */
+  mess = daemon->dhcp_packet.iov_base;
+  
 #ifdef HAVE_SOCKADDR_SA_LEN
-      dest.sin_len = sizeof(struct sockaddr_in);
+  dest.sin_len = sizeof(struct sockaddr_in);
 #endif
      
-      if (mess->giaddr.s_addr)
-	{
-	  dest.sin_port = htons(DHCP_SERVER_PORT);
-	  dest.sin_addr = mess->giaddr; 
-	}
-      else
-	{
-	  dest.sin_port = htons(DHCP_CLIENT_PORT);
-	  dest.sin_addr = mess->ciaddr;
-	}
-      
-      while(sendto(daemon->dhcpfd, mess, newlen, 0, 
-		   (struct sockaddr *)&dest, sizeof(dest)) == -1 &&
-	    retry_send());
+  if (mess->giaddr.s_addr)
+    {
+      /* Send to BOOTP relay  */
+      if (!dest.sin_port)
+	dest.sin_port = htons(DHCP_SERVER_PORT);
+      dest.sin_addr = mess->giaddr; 
+    }
+  else if (mess->ciaddr.s_addr)
+    {
+      dest.sin_addr = mess->ciaddr;
+      if (!dest.sin_port)
+	dest.sin_port = htons(DHCP_CLIENT_PORT);
+    } 
+#ifdef HAVE_LINUX_NETWORK
+  else if (ntohs(mess->flags) & 0x8000)
+    {
+      /* broadcast to 255.255.255.255 */
+      struct in_pktinfo *pkt;
+      msg.msg_control = control_u.control;
+      msg.msg_controllen = sizeof(control_u);
+      cmptr = CMSG_FIRSTHDR(&msg);
+      dest.sin_addr.s_addr = INADDR_BROADCAST;
+      dest.sin_port = htons(DHCP_CLIENT_PORT);
+      pkt = (struct in_pktinfo *)CMSG_DATA(cmptr);
+      pkt->ipi_ifindex = iface_index;
+      pkt->ipi_spec_dst.s_addr = 0;
+      msg.msg_controllen = cmptr->cmsg_len = CMSG_LEN(sizeof(struct in_pktinfo));
+      cmptr->cmsg_level = SOL_IP;
+      cmptr->cmsg_type = IP_PKTINFO;
     }
   else
     {
-      /* Hairy stuff, packet either has to go to the
-	 net broadcast or the destination can't reply to ARP yet,
-	 but we do know the physical address. 
-	 Build the packet by steam, and send directly, bypassing
-	 the kernel IP stack */
-      
-      u32 i, sum;
-      unsigned char hwdest[ETHER_ADDR_LEN];
-      
-      if (ntohs(mess->flags) & 0x8000)
+      /* unicast to unconfigured client */
+      dest.sin_addr = mess->yiaddr;
+      dest.sin_port = htons(DHCP_CLIENT_PORT);
+      if (mess->hlen != 0 && mess->hlen <= 14 && mess->htype != 0)
 	{
-	  memset(hwdest, 255,  ETHER_ADDR_LEN);
-	  rawpacket->ip.ip_dst.s_addr = INADDR_BROADCAST;
+	  /* inject mac address direct into ARP cache. 
+	     struct sockaddr limits size to 14 bytes. */
+	  struct arpreq req;
+	  *((struct sockaddr_in *)&req.arp_pa) = dest;
+	  req.arp_ha.sa_family = mess->htype;
+	  memcpy(req.arp_ha.sa_data, mess->chaddr, mess->hlen);
+	  strncpy(req.arp_dev, ifr.ifr_name, 16);
+	  req.arp_flags = ATF_COM;
+	  ioctl(daemon->dhcpfd, SIOCSARP, &req);
 	}
-      else
-	{
-	  memcpy(hwdest, mess->chaddr, ETHER_ADDR_LEN); 
-	  rawpacket->ip.ip_dst.s_addr = mess->yiaddr.s_addr;
-	}
-      
-      rawpacket->ip.ip_p = IPPROTO_UDP;
-      rawpacket->ip.ip_src.s_addr = iface_addr.s_addr;
-      rawpacket->ip.ip_len = htons(sizeof(struct ip) + 
-				   sizeof(struct udphdr) +
-				   newlen) ;
-      rawpacket->ip.ip_hl = sizeof(struct ip) / 4;
-      rawpacket->ip.ip_v = IPVERSION;
-      rawpacket->ip.ip_tos = 0;
-      rawpacket->ip.ip_id = htons(0);
-      rawpacket->ip.ip_off = htons(0x4000); /* don't fragment */
-      rawpacket->ip.ip_ttl = IPDEFTTL;
-      rawpacket->ip.ip_sum = 0;
-      for (sum = 0, i = 0; i < sizeof(struct ip) / 2; i++)
-	sum += ((u16 *)&rawpacket->ip)[i];
-      while (sum>>16)
-	sum = (sum & 0xffff) + (sum >> 16);  
-      rawpacket->ip.ip_sum = (sum == 0xffff) ? sum : ~sum;
-      
-      rawpacket->udp.uh_sport = htons(DHCP_SERVER_PORT);
-      rawpacket->udp.uh_dport = htons(DHCP_CLIENT_PORT);
-      ((u8 *)&rawpacket->data)[newlen] = 0; /* for checksum, in case length is odd. */
-      rawpacket->udp.uh_sum = 0;
-      rawpacket->udp.uh_ulen = sum = htons(sizeof(struct udphdr) + newlen);
-      sum += htons(IPPROTO_UDP);
-      for (i = 0; i < 4; i++)
-	sum += ((u16 *)&rawpacket->ip.ip_src)[i];
-      for (i = 0; i < (sizeof(struct udphdr) + newlen + 1) / 2; i++)
-	sum += ((u16 *)&rawpacket->udp)[i];
-      while (sum>>16)
-	sum = (sum & 0xffff) + (sum >> 16);
-      rawpacket->udp.uh_sum = (sum == 0xffff) ? sum : ~sum;
-      
-      { 
-#ifdef HAVE_BPF
-	struct ether_header header;
-	
-	header.ether_type = htons(ETHERTYPE_IP);
-	memcpy(header.ether_shost, iface_hwaddr, ETHER_ADDR_LEN);
-	memcpy(header.ether_dhost, hwdest, ETHER_ADDR_LEN); 
-	
-	ioctl(daemon->dhcp_raw_fd, BIOCSETIF, &ifr);
-	
-	iov[0].iov_base = (char *)&header;
-	iov[0].iov_len = sizeof(struct ether_header);
-	iov[1].iov_base = (char *)rawpacket;
-	iov[1].iov_len = ntohs(rawpacket->ip.ip_len);
-	while (writev(daemon->dhcp_raw_fd, iov, 2) == -1 && retry_send());
-#else
-	struct sockaddr_ll dest;
-	
-	memset(&dest, 0, sizeof(dest));
-	dest.sll_family = AF_PACKET;
-	dest.sll_halen =  ETHER_ADDR_LEN;
-	dest.sll_ifindex = iface_index;
-	dest.sll_protocol = htons(ETHERTYPE_IP);
-	memcpy(dest.sll_addr, hwdest, ETHER_ADDR_LEN); 
-	while (sendto(daemon->dhcp_raw_fd, rawpacket, ntohs(rawpacket->ip.ip_len), 
-		      0, (struct sockaddr *)&dest, sizeof(dest)) == -1 &&
-	       retry_send());
-#endif
-      }
     }
+#else
+  else 
+    {
+      send_via_bpf(daemon, mess, iov.iov_len, iface_addr, &ifr);
+      return;
+    }
+#endif
+   
+  while(sendmsg(daemon->dhcpfd, &msg, 0) == -1 && retry_send());
 }
-
+ 
 /* This is a complex routine: it gets called with each (address,netmask,broadcast) triple 
-   of the interface on which a DHCP packet arrives (and any relay address) and does the 
-   following things:
-   1) Fills in any netmask and broadcast addresses which have not been explicitly configured.
-   2) Fills in local (this host) and router (this host or relay) addresses.
-   3) Links contexts which are valid for hosts directly connected to the arrival interface on ->current.
+   of each interface (and any relay address) and does the  following things:
+
+   1) Discards stuff for interfaces other than the one on which a DHCP packet just arrived.
+   2) Fills in any netmask and broadcast addresses which have not been explicitly configured.
+   3) Fills in local (this host) and router (this host or relay) addresses.
+   4) Links contexts which are valid for hosts directly connected to the arrival interface on ->current.
+
    Note that the current chain may be superceded later for configured hosts or those coming via gateways. */
-struct dhcp_context *complete_context(struct daemon *daemon, struct in_addr local, struct dhcp_context *current,
-				      struct in_addr netmask, struct in_addr broadcast, struct in_addr relay,
-				      struct in_addr primary)
+
+static int complete_context(struct daemon *daemon, struct in_addr local, int if_index, 
+			     struct in_addr netmask, struct in_addr broadcast, void *vparam)
 {
   struct dhcp_context *context;
+  struct iface_param *param = vparam;
+  
+  if (if_index != param->ind)
+    return 1; /* no for us. */
   
   for (context = daemon->dhcp; context; context = context->next)
     {
@@ -376,7 +313,7 @@ struct dhcp_context *complete_context(struct daemon *daemon, struct in_addr loca
 	  {
 	    strcpy(daemon->dhcp_buff, inet_ntoa(context->start));
 	    strcpy(daemon->dhcp_buff2, inet_ntoa(context->end));
-	    syslog(LOG_WARNING, "DHCP range %s -- %s is not consistent with netmask %s",
+	    syslog(LOG_WARNING, _("DHCP range %s -- %s is not consistent with netmask %s"),
 		   daemon->dhcp_buff, daemon->dhcp_buff2, inet_ntoa(netmask));
 	  }	
  	context->netmask = netmask;
@@ -392,8 +329,8 @@ struct dhcp_context *complete_context(struct daemon *daemon, struct in_addr loca
 		{
 		  context->router = local;
 		  context->local = local;
-		  context->current = current;
-		  current = context;
+		  context->current = param->current;
+		  param->current = context;
 		}
 	      
 	      if (!(context->flags & CONTEXT_BRDCAST))
@@ -404,10 +341,10 @@ struct dhcp_context *complete_context(struct daemon *daemon, struct in_addr loca
 		    context->broadcast.s_addr  = context->start.s_addr | ~context->netmask.s_addr;
 		}
 	    }	
-	  else if (relay.s_addr && is_same_net(relay, context->start, context->netmask))
+	  else if (param->relay.s_addr && is_same_net(param->relay, context->start, context->netmask))
 	    {
-	      context->router = relay;
-	      context->local = primary;
+	      context->router = param->relay;
+	      context->local = param->primary;
 	      /* fill in missing broadcast addresses for relayed ranges */
 	      if (!(context->flags & CONTEXT_BRDCAST))
 		context->broadcast.s_addr  = context->start.s_addr | ~context->netmask.s_addr;
@@ -416,25 +353,31 @@ struct dhcp_context *complete_context(struct daemon *daemon, struct in_addr loca
 	}
     }
 
-  return current;
+  return 1;
 }
 	  
 struct dhcp_context *address_available(struct dhcp_context *context, struct in_addr taddr)
 {
   /* Check is an address is OK for this network, check all
-     possible ranges. */
+     possible ranges. Make sure that the address isn't in use
+     by the server itself. */
   
   unsigned int start, end, addr = ntohl(taddr.s_addr);
-  
-  for (; context; context = context->current)
-    {
-      start = ntohl(context->start.s_addr);
-      end = ntohl(context->end.s_addr);
+  struct dhcp_context *tmp;
 
-      if (!(context->flags & CONTEXT_STATIC) &&
+  for (tmp = context; tmp; tmp = tmp->current)
+    if (taddr.s_addr == context->local.s_addr)
+      return NULL;
+  
+  for (tmp = context; tmp; tmp = tmp->current)
+    {
+      start = ntohl(tmp->start.s_addr);
+      end = ntohl(tmp->end.s_addr);
+
+      if (!(tmp->flags & CONTEXT_STATIC) &&
 	  addr >= start &&
 	  addr <= end)
-	return context;
+	return tmp;
     }
 
   return NULL;
@@ -442,24 +385,29 @@ struct dhcp_context *address_available(struct dhcp_context *context, struct in_a
 
 struct dhcp_context *narrow_context(struct dhcp_context *context, struct in_addr taddr)
 {
-  /* We start of with a set of possible contexts, all on the current subnet.
+  /* We start of with a set of possible contexts, all on the current physical interface.
      These are chained on ->current.
      Here we have an address, and return the actual context correponding to that
      address. Note that none may fit, if the address came a dhcp-host and is outside
-     any dhcp-range. In that case we return a static range is possible, or failing that,
-     any context on the subnet. (If there's more than one, this is a dodgy configuration: 
-     maybe there should be a warning.) */
+     any dhcp-range. In that case we return a static range if possible, or failing that,
+     any context on the correct subnet. (If there's more than one, this is a dodgy 
+     configuration: maybe there should be a warning.) */
   
-  struct dhcp_context *tmp = address_available(context, taddr);
+  struct dhcp_context *tmp;
 
-  if (tmp)
+  if ((tmp = address_available(context, taddr)))
     return tmp;
   
   for (tmp = context; tmp; tmp = tmp->current)
-    if (tmp->flags & CONTEXT_STATIC)
+    if (is_same_net(taddr, tmp->start, tmp->netmask) && 
+	(tmp->flags & CONTEXT_STATIC))
       return tmp;
 
-  return context;
+  for (tmp = context; tmp; tmp = tmp->current)
+    if (is_same_net(taddr, tmp->start, tmp->netmask))
+      return tmp;
+
+  return NULL;
 }
 
 struct dhcp_config *config_find_by_address(struct dhcp_config *configs, struct in_addr addr)
@@ -473,12 +421,13 @@ struct dhcp_config *config_find_by_address(struct dhcp_config *configs, struct i
   return NULL;
 }
 
-/* Is every member of check matched by a member of pool? */
-int match_netid(struct dhcp_netid *check, struct dhcp_netid *pool)
+/* Is every member of check matched by a member of pool? 
+   If negonly, match unless there's a negative tag which matches. */
+int match_netid(struct dhcp_netid *check, struct dhcp_netid *pool, int negonly)
 {
   struct dhcp_netid *tmp1;
   
-  if (!check)
+  if (!check && !negonly)
     return 0;
 
   for (; check; check = check->next)
@@ -488,7 +437,7 @@ int match_netid(struct dhcp_netid *check, struct dhcp_netid *pool)
 	  for (tmp1 = pool; tmp1; tmp1 = tmp1->next)
 	    if (strcmp(check->net, tmp1->net) == 0)
 	      break;
-	  if (!tmp1)
+	  if (!tmp1 || negonly)
 	    return 0;
 	}
       else
@@ -500,108 +449,112 @@ int match_netid(struct dhcp_netid *check, struct dhcp_netid *pool)
 }
 
 int address_allocate(struct dhcp_context *context, struct daemon *daemon,
-		     struct in_addr *addrp, unsigned char *hwaddr, 
+		     struct in_addr *addrp, unsigned char *hwaddr, int hw_len, 
 		     struct dhcp_netid *netids, time_t now)   
 {
   /* Find a free address: exclude anything in use and anything allocated to
      a particular hwaddr/clientid/hostname in our configuration.
-     Try to return from contexts which mathc netis first. */
+     Try to return from contexts which match netids first. */
 
-  struct in_addr start, addr ;
-  struct dhcp_context *c;
-  unsigned int i, j;
+  struct in_addr start, addr;
+  struct dhcp_context *c, *d;
+  int i, pass;
+  unsigned int j; 
+
+  /* hash hwaddr */
+  for (j = 0, i = 0; i < hw_len; i++)
+    j += hwaddr[i] + (hwaddr[i] << 8) + (hwaddr[i] << 16);
   
-  for (c = context; c; c = c->current)
-    if (c->flags & CONTEXT_STATIC)
-      continue;
-    else if (netids && !(c->flags & CONTEXT_FILTER))
-      continue;
-    else if (!netids && (c->flags & CONTEXT_FILTER))
-      continue;
-    else if (netids && (c->flags & CONTEXT_FILTER) && !match_netid(&c->netid, netids))
-      continue;
-    else
-      {
-	/* pick a seed based on hwaddr then iterate until we find a free address. */
-	for (j = c->addr_epoch, i = 0; i < ETHER_ADDR_LEN; i++)
-	  j += hwaddr[i] + (hwaddr[i] << 8) + (hwaddr[i] << 16);
-	
-	start.s_addr = addr.s_addr = 
-	  htonl(ntohl(c->start.s_addr) + 
-		(j % (1 + ntohl(c->end.s_addr) - ntohl(c->start.s_addr))));
-	
-	do {
-	  if (!lease_find_by_addr(addr) && 
-	      !config_find_by_address(daemon->dhcp_conf, addr))
-	    {
-	      struct ping_result *r, *victim = NULL;
-	      int count;
+  for (pass = 0; pass <= 1; pass++)
+    for (c = context; c; c = c->current)
+      if (c->flags & CONTEXT_STATIC)
+	continue;
+      else if (!match_netid(c->filter, netids, pass))
+	continue;
+      else
+	{
+	  /* pick a seed based on hwaddr then iterate until we find a free address. */
+	  start.s_addr = addr.s_addr = 
+	    htonl(ntohl(c->start.s_addr) + 
+		  ((j + c->addr_epoch) % (1 + ntohl(c->end.s_addr) - ntohl(c->start.s_addr))));
+	  
+	  do {
+	    /* eliminate addresses in use by the server. */
+	    for (d = context; d; d = d->current)
+	      if (addr.s_addr == d->local.s_addr)
+		break;
 
-	      /* check if we failed to ping addr sometime in the last
-		 30s. If so, assume the same situation still exists.
-		 This avoids problems when a stupid client bangs
-		 on us repeatedly. As a final check, is we did more
-		 than six ping checks in the last 30s, we are in 
-		 high-load mode, so don't do any more. */
-	      for (count = 0, r = daemon->ping_results; r; r = r->next)
-		if (difftime(now, r->time) > 30.0)
-		  victim = r; /* old record */
-		else if (++count == 6 || r->addr.s_addr == addr.s_addr)
+	    if (!d &&
+		!lease_find_by_addr(addr) && 
+		!config_find_by_address(daemon->dhcp_conf, addr))
+	      {
+		struct ping_result *r, *victim = NULL;
+		int count, max = (int)(0.6 * (((float)PING_CACHE_TIME)/
+					      ((float)PING_WAIT)));
+		
+		*addrp = addr;
+
+		if (daemon->options & OPT_NO_PING)
+		  return 1;
+		
+		/* check if we failed to ping addr sometime in the last
+		   PING_CACHE_TIME seconds. If so, assume the same situation still exists.
+		   This avoids problems when a stupid client bangs
+		   on us repeatedly. As a final check, if we did more
+		   than 60% of the possible ping checks in the last 
+		   PING_CACHE_TIME, we are in high-load mode, so don't do any more. */
+		for (count = 0, r = daemon->ping_results; r; r = r->next)
+		  if (difftime(now, r->time) >  (float)PING_CACHE_TIME)
+		    victim = r; /* old record */
+		  else if (++count == max || r->addr.s_addr == addr.s_addr)
+		    return 1;
+		    
+		if (icmp_ping(daemon, addr))
+		  /* address in use: perturb address selection so that we are
+		     less likely to try this address again. */
+		  c->addr_epoch++;
+		else
 		  {
-		    *addrp = addr;
+		    /* at this point victim may hold an expired record */
+		    if (!victim)
+		      {
+			if ((victim = malloc(sizeof(struct ping_result))))
+			  {
+			    victim->next = daemon->ping_results;
+			    daemon->ping_results = victim;
+			  }
+		      }
+		    
+		    /* record that this address is OK for 30s 
+		       without more ping checks */
+		    if (victim)
+		      {
+			victim->addr = addr;
+			victim->time = now;
+		      }
 		    return 1;
 		  }
-	      
-	      if (icmp_ping(daemon, addr))
-		/* address in use: perturb address selection so that we are
-		   less likely to try this address again. */
-		c->addr_epoch++;
-	      else
-		{
-		  /* at this point victim may hold an expired record */
-		  if (!victim)
-		    {
-		      if ((victim = malloc(sizeof(struct ping_result))))
-			{
-			  victim->next = daemon->ping_results;
-			  daemon->ping_results = victim;
-			}
-		    }
-		  
-		  /* record that this address is OK for 30s 
-		     without more ping checks */
-		  if (victim)
-		    {
-		      victim->addr = addr;
-		      victim->time = now;
-		    }
-		  *addrp = addr;
-		  return 1;
-		}
-	    }
-
-	  addr.s_addr = htonl(ntohl(addr.s_addr) + 1);
-	  
-	  if (addr.s_addr == htonl(ntohl(c->end.s_addr) + 1))
-	    addr = c->start;
-	  
-	} while (addr.s_addr != start.s_addr);
-      }
-
-  if (netids)
-    return address_allocate(context, daemon, addrp, hwaddr, NULL, now);
+	      }
 
+	    addr.s_addr = htonl(ntohl(addr.s_addr) + 1);
+	    
+	    if (addr.s_addr == htonl(ntohl(c->end.s_addr) + 1))
+	      addr = c->start;
+	    
+	  } while (addr.s_addr != start.s_addr);
+	}
   return 0;
 }
 
 static int is_addr_in_context(struct dhcp_context *context, struct dhcp_config *config)
 {
-  if (!context)
+  if (!context) /* called via find_config() from lease_update_from_configs() */
     return 1; 
   if (!(config->flags & CONFIG_ADDR))
     return 1;
-  if (is_same_net(config->addr, context->start, context->netmask))
-    return 1;
+  for (; context; context = context->current)
+    if (is_same_net(config->addr, context->start, context->netmask))
+      return 1;
   
   return 0;
 }
@@ -610,7 +563,8 @@ static int is_addr_in_context(struct dhcp_context *context, struct dhcp_config *
 struct dhcp_config *find_config(struct dhcp_config *configs,
 				struct dhcp_context *context,
 				unsigned char *clid, int clid_len,
-				unsigned char *hwaddr, char *hostname)
+				unsigned char *hwaddr, int hw_len, 
+				int hw_type, char *hostname)
 {
   struct dhcp_config *config; 
   
@@ -632,14 +586,14 @@ struct dhcp_config *find_config(struct dhcp_config *configs,
 	}
   
 
-  if (hwaddr)
-    for (config = configs; config; config = config->next)
-      if ((config->flags & CONFIG_HWADDR) &&
-	  config->wildcard_mask == 0 &&
-	  memcmp(config->hwaddr, hwaddr, ETHER_ADDR_LEN) == 0 &&
-	  is_addr_in_context(context, config))
-	return config;
-  
+  for (config = configs; config; config = config->next)
+    if ((config->flags & CONFIG_HWADDR) &&
+	config->wildcard_mask == 0 &&
+	config->hwaddr_len == hw_len &&
+	(config->hwaddr_type == hw_type || config->hwaddr_type == 0) &&
+	memcmp(config->hwaddr, hwaddr, hw_len) == 0 &&
+	is_addr_in_context(context, config))
+      return config;
   
   if (hostname && context)
     for (config = configs; config; config = config->next)
@@ -648,21 +602,15 @@ struct dhcp_config *find_config(struct dhcp_config *configs,
 	  is_addr_in_context(context, config))
 	return config;
   
-  if (hwaddr)
-    for (config = configs; config; config = config->next)
-      if ((config->flags & CONFIG_HWADDR) &&
-	  config->wildcard_mask != 0 &&
-	  is_addr_in_context(context, config))
-	{
-	  int i;
-	  unsigned int mask = config->wildcard_mask;
-	  for (i = ETHER_ADDR_LEN - 1; i >= 0; i--, mask = mask >> 1)
-	    if (mask & 1)
-	      config->hwaddr[i] = hwaddr[i];
-	  if (memcmp(config->hwaddr, hwaddr, ETHER_ADDR_LEN) == 0)
-	    return config;
-	}
-  
+  for (config = configs; config; config = config->next)
+    if ((config->flags & CONFIG_HWADDR) &&
+	config->wildcard_mask != 0 &&
+	config->hwaddr_len == hw_len &&	
+	(config->hwaddr_type == hw_type || config->hwaddr_type == 0) &&
+	is_addr_in_context(context, config) &&
+	memcmp_masked(config->hwaddr, hwaddr, hw_len, config->wildcard_mask))
+      return config;
+        
   return NULL;
 }
 
@@ -675,18 +623,20 @@ void dhcp_read_ethers(struct daemon *daemon)
   struct in_addr addr;
   unsigned char hwaddr[ETHER_ADDR_LEN];
   struct dhcp_config *config, *configs = daemon->dhcp_conf;
-  int count = 0;
+  int count = 0, lineno = 0;
 
   addr.s_addr = 0; /* eliminate warning */
   
   if (!f)
     {
-      syslog(LOG_ERR, "failed to read " ETHERSFILE ":%m");
+      syslog(LOG_ERR, _("failed to read %s:%m"), ETHERSFILE);
       return;
     }
 
   while (fgets(buff, MAXDNAME, f))
     {
+      lineno++;
+      
       while (strlen(buff) > 0 && isspace(buff[strlen(buff)-1]))
 	buff[strlen(buff)-1] = 0;
       
@@ -696,11 +646,11 @@ void dhcp_read_ethers(struct daemon *daemon)
       for (ip = buff; *ip && !isspace(*ip); ip++);
       for(; *ip && isspace(*ip); ip++)
 	*ip = 0;
-      if (!*ip)
-	continue;
-      
-      if (parse_hex(buff, hwaddr, 6, NULL) != 6)
-	continue;
+      if (!*ip || parse_hex(buff, hwaddr, ETHER_ADDR_LEN, NULL, NULL) != ETHER_ADDR_LEN)
+	{
+	  syslog(LOG_ERR, _("bad line at %s line %d"), ETHERSFILE, lineno); 
+	  continue;
+	}
       
       /* check for name or dotted-quad */
       for (cp = ip; *cp; cp++)
@@ -710,7 +660,11 @@ void dhcp_read_ethers(struct daemon *daemon)
       if (!*cp)
 	{
 	  if ((addr.s_addr = inet_addr(ip)) == (in_addr_t)-1)
-	    continue;
+	    {
+	      syslog(LOG_ERR, _("bad address at %s line %d"), ETHERSFILE, lineno); 
+	      continue;
+	    }
+
 	  flags = CONFIG_ADDR;
 	  
 	  for (config = configs; config; config = config->next)
@@ -720,7 +674,11 @@ void dhcp_read_ethers(struct daemon *daemon)
       else 
 	{
 	  if (!canonicalise(ip))
-	    continue;
+	    {
+	      syslog(LOG_ERR, _("bad name at %s line %d"), ETHERSFILE, lineno); 
+	      continue;
+	    }
+
 	  flags = CONFIG_NAME;
 
 	  for (config = configs; config; config = config->next)
@@ -733,6 +691,8 @@ void dhcp_read_ethers(struct daemon *daemon)
 	  for (config = configs; config; config = config->next)
 	    if ((config->flags & CONFIG_HWADDR) && 
 		config->wildcard_mask == 0 &&
+		config->hwaddr_len == ETHER_ADDR_LEN &&
+		(config->hwaddr_type == ARPHRD_ETHER || config->hwaddr_type == 0) &&
 		memcmp(config->hwaddr, hwaddr, ETHER_ADDR_LEN) == 0)
 	      break;
 	  
@@ -762,13 +722,14 @@ void dhcp_read_ethers(struct daemon *daemon)
       
       config->flags |= CONFIG_HWADDR | CONFIG_NOCLID;
       memcpy(config->hwaddr, hwaddr, ETHER_ADDR_LEN);
-
+      config->hwaddr_len = ETHER_ADDR_LEN;
+      config->hwaddr_type = ARPHRD_ETHER;
       count++;
     }
   
   fclose(f);
 
-  syslog(LOG_INFO, "read " ETHERSFILE " - %d addresses", count);
+  syslog(LOG_INFO, _("read %s - %d addresses"), ETHERSFILE, count);
   
   daemon->dhcp_conf =  configs;
 }
@@ -791,7 +752,7 @@ void dhcp_update_configs(struct dhcp_config *configs)
 	(crec->flags & F_HOSTS))
       {
 	if (config_find_by_address(configs, crec->addr.addr.addr.addr4))
-	  syslog(LOG_WARNING, "duplicate IP address %s (%s) in dhcp-config directive", 
+	  syslog(LOG_WARNING, _("duplicate IP address %s (%s) in dhcp-config directive"), 
 		 inet_ntoa(crec->addr.addr.addr.addr4), config->hostname);
 	else
 	  {
@@ -812,8 +773,8 @@ char *host_from_dns(struct daemon *daemon, struct in_addr addr)
   if (lookup && (lookup->flags & F_HOSTS))
     {
       hostname = daemon->dhcp_buff;
-      hostname[256] = 0;
       strncpy(hostname, cache_get_name(lookup), 256);
+      hostname[255] = 0;
       hostname = strip_hostname(daemon, hostname);
     }
 
@@ -827,7 +788,7 @@ char *strip_hostname(struct daemon *daemon, char *hostname)
     {
       if (!daemon->domain_suffix || !hostname_isequal(dot+1, daemon->domain_suffix))
 	{
-	  syslog(LOG_WARNING, "Ignoring DHCP host name %s because it has an illegal domain part", hostname);
+	  syslog(LOG_WARNING, _("Ignoring DHCP host name %s because it has an illegal domain part"), hostname);
 	  hostname = NULL;
 	}
       else
