@@ -10,8 +10,6 @@
    GNU General Public License for more details.
 */
 
-/* Author's email: simon@thekelleys.org.uk */
-
 #include "dnsmasq.h"
 
 static struct frec *frec_list;
@@ -36,7 +34,7 @@ void forward_init(int first)
 
 /* Send a UDP packet with it's source address set as "source" 
    unless nowild is true, when we just send it with the kernel default */
-static void send_from(int fd, int nowild, char *packet, int len, 
+static void send_from(int fd, int nowild, char *packet, size_t len, 
 		      union mysockaddr *to, struct all_addr *source,
 		      unsigned int iface)
 {
@@ -44,7 +42,7 @@ static void send_from(int fd, int nowild, char *packet, int len,
   struct iovec iov[1]; 
   union {
     struct cmsghdr align; /* this ensures alignment */
-#if defined(IP_PKTINFO)
+#if defined(HAVE_LINUX_NETWORK)
     char control[CMSG_SPACE(sizeof(struct in_pktinfo))];
 #elif defined(IP_SENDSRCADDR)
     char control[CMSG_SPACE(sizeof(struct in_addr))];
@@ -74,7 +72,7 @@ static void send_from(int fd, int nowild, char *packet, int len,
 
       if (to->sa.sa_family == AF_INET)
 	{
-#if defined(IP_PKTINFO)
+#if defined(HAVE_LINUX_NETWORK)
 	  struct in_pktinfo *pkt = (struct in_pktinfo *)CMSG_DATA(cmptr);
 	  pkt->ipi_ifindex = 0;
 	  pkt->ipi_spec_dst = source->addr.addr4;
@@ -89,9 +87,8 @@ static void send_from(int fd, int nowild, char *packet, int len,
 	  cmptr->cmsg_type = IP_SENDSRCADDR;
 #endif
 	}
-      
-#ifdef HAVE_IPV6
       else
+#ifdef HAVE_IPV6
 	{
 	  struct in6_pktinfo *pkt = (struct in6_pktinfo *)CMSG_DATA(cmptr);
 	  pkt->ipi6_ifindex = iface; /* Need iface for IPv6 to handle link-local addrs */
@@ -100,6 +97,8 @@ static void send_from(int fd, int nowild, char *packet, int len,
 	  cmptr->cmsg_type = IPV6_PKTINFO;
 	  cmptr->cmsg_level = IPV6_LEVEL;
 	}
+#else
+      iface = 0; /* eliminate warning */
 #endif
     }
   
@@ -134,7 +133,7 @@ static unsigned short search_servers(struct daemon *daemon, time_t now, struct a
   
   for (serv = daemon->servers; serv; serv=serv->next)
     /* domain matches take priority over NODOTS matches */
-    if ((serv->flags & SERV_FOR_NODOTS) && *type != SERV_HAS_DOMAIN && !strchr(qdomain, '.'))
+    if ((serv->flags & SERV_FOR_NODOTS) && *type != SERV_HAS_DOMAIN && !strchr(qdomain, '.') && namelen != 0)
       {
 	unsigned short sflag = serv->addr.sa.sa_family == AF_INET ? F_IPV4 : F_IPV6; 
 	*type = SERV_FOR_NODOTS;
@@ -159,9 +158,11 @@ static unsigned short search_servers(struct daemon *daemon, time_t now, struct a
     else if (serv->flags & SERV_HAS_DOMAIN)
       {
 	unsigned int domainlen = strlen(serv->domain);
+	char *matchstart = qdomain + namelen - domainlen;
 	if (namelen >= domainlen &&
-	    hostname_isequal(qdomain + namelen - domainlen, serv->domain) &&
-	    domainlen >= matchlen)
+	    hostname_isequal(matchstart, serv->domain) &&
+	    domainlen >= matchlen &&
+	    (domainlen == 0 || namelen == domainlen || *(serv->domain) == '.' || *(matchstart-1) == '.' ))
 	  {
 	    unsigned short sflag = serv->addr.sa.sa_family == AF_INET ? F_IPV4 : F_IPV6;
 	    *type = SERV_HAS_DOMAIN;
@@ -194,7 +195,7 @@ static unsigned short search_servers(struct daemon *daemon, time_t now, struct a
       else
 	log_query(F_CONFIG | F_FORWARD | flags, qdomain, *addrpp, 0, NULL, 0);
     }
-  else if (qtype && (daemon->options & OPT_NODOTS_LOCAL) && !strchr(qdomain, '.'))
+  else if (qtype && (daemon->options & OPT_NODOTS_LOCAL) && !strchr(qdomain, '.') && namelen != 0)
     flags = F_NXDOMAIN;
     
   if (flags == F_NXDOMAIN && check_for_local_domain(qdomain, now, daemon))
@@ -209,21 +210,20 @@ static unsigned short search_servers(struct daemon *daemon, time_t now, struct a
 /* returns new last_server */	
 static void forward_query(struct daemon *daemon, int udpfd, union mysockaddr *udpaddr,
 			  struct all_addr *dst_addr, unsigned int dst_iface,
-			  HEADER *header, int plen, time_t now)
+			  HEADER *header, size_t plen, time_t now, struct frec *forward)
 {
-  struct frec *forward;
   char *domain = NULL;
   int type = 0;
   struct all_addr *addrp = NULL;
+  unsigned int crc = questions_crc(header, plen, daemon->namebuff);
   unsigned short flags = 0;
-  unsigned short gotname = extract_request(header, (unsigned int)plen, daemon->namebuff, NULL);
+  unsigned short gotname = extract_request(header, plen, daemon->namebuff, NULL);
   struct server *start = NULL;
-  unsigned int crc = questions_crc(header,(unsigned int)plen, daemon->namebuff);
-  
-  /* may be  recursion not speced or no servers available. */
-  if (!header->rd || !daemon->servers)
+    
+  /* may be no servers available. */
+  if (!daemon->servers)
     forward = NULL;
-  else if ((forward = lookup_frec_by_sender(ntohs(header->id), udpaddr, crc)))
+  else if (forward || (forward = lookup_frec_by_sender(ntohs(header->id), udpaddr, crc)))
     {
       /* retry on existing query, send to all available servers  */
       domain = forward->sentto->domain;
@@ -300,6 +300,10 @@ static void forward_query(struct daemon *daemon, int udpfd, union mysockaddr *ud
 		}
 	      else
 		{
+		  /* Keep info in case we want to re-send this packet */
+		  daemon->srv_save = start;
+		  daemon->packet_len = plen;
+		  
 		  if (!gotname)
 		    strcpy(daemon->namebuff, "query");
 		  if (start->addr.sa.sa_family == AF_INET)
@@ -336,18 +340,22 @@ static void forward_query(struct daemon *daemon, int udpfd, union mysockaddr *ud
     }	  
   
   /* could not send on, return empty answer or address if known for whole domain */
-  plen = setup_reply(header, (unsigned int)plen, addrp, flags, daemon->local_ttl);
-  send_from(udpfd, daemon->options & OPT_NOWILD, (char *)header, plen, udpaddr, dst_addr, dst_iface);
-  
+  if (udpfd != -1)
+    {
+      plen = setup_reply(header, plen, addrp, flags, daemon->local_ttl);
+      send_from(udpfd, daemon->options & OPT_NOWILD, (char *)header, plen, udpaddr, dst_addr, dst_iface);
+    }
+
   return;
 }
 
-static int process_reply(struct daemon *daemon, HEADER *header, time_t now, 
-			 unsigned int query_crc, struct server *server, unsigned int n)
+static size_t process_reply(struct daemon *daemon, HEADER *header, time_t now, 
+			    unsigned int query_crc, struct server *server, size_t n)
 {
   unsigned char *pheader, *sizep;
-  unsigned int plen, munged = 0;
-   
+  int munged = 0;
+  size_t plen; 
+
   /* If upstream is advertising a larger UDP packet size
 	 than we allow, trim it so that we don't get overlarge
 	 requests for the client. */
@@ -369,16 +377,8 @@ static int process_reply(struct daemon *daemon, HEADER *header, time_t now,
   if (!header->ra && header->rcode == NOERROR && ntohs(header->ancount) == 0 &&
       server && !(server->flags & SERV_WARNED_RECURSIVE))
     {
-      char addrbuff[ADDRSTRLEN];
-#ifdef HAVE_IPV6
-      if (server->addr.sa.sa_family == AF_INET)
-	inet_ntop(AF_INET, &server->addr.in.sin_addr, addrbuff, ADDRSTRLEN);
-      else if (server->addr.sa.sa_family == AF_INET6)
-	inet_ntop(AF_INET6, &server->addr.in6.sin6_addr, addrbuff, ADDRSTRLEN);
-#else
-      strcpy(addrbuff, inet_ntoa(server->addr.in.sin_addr));
-#endif
-      syslog(LOG_WARNING, "nameserver %s refused to do a recursive query", addrbuff);
+      prettyprint_addr(&server->addr, daemon->namebuff);
+      syslog(LOG_WARNING, _("nameserver %s refused to do a recursive query"), daemon->namebuff);
       if (!(daemon->options & OPT_LOG))
 	server->flags |= SERV_WARNED_RECURSIVE;
     }  
@@ -436,13 +436,17 @@ void reply_query(struct serverfd *sfd, struct daemon *daemon, time_t now)
   HEADER *header;
   union mysockaddr serveraddr;
   socklen_t addrlen = sizeof(serveraddr);
-  int n = recvfrom(sfd->fd, daemon->packet, daemon->edns_pktsz, 0, &serveraddr.sa, &addrlen);
-  
+  ssize_t n = recvfrom(sfd->fd, daemon->packet, daemon->edns_pktsz, 0, &serveraddr.sa, &addrlen);
+  size_t nn;
+
+  /* packet buffer overwritten */
+  daemon->srv_save = NULL;
+
   /* Determine the address of the server replying  so that we can mark that as good */
   serveraddr.sa.sa_family = sfd->source_addr.sa.sa_family;
 #ifdef HAVE_IPV6
   if (serveraddr.sa.sa_family == AF_INET6)
-    serveraddr.in6.sin6_flowinfo = htonl(0);
+    serveraddr.in6.sin6_flowinfo = 0;
 #endif
   
   header = (HEADER *)daemon->packet;
@@ -452,40 +456,62 @@ void reply_query(struct serverfd *sfd, struct daemon *daemon, time_t now)
     {
        struct server *server = forward->sentto;
        
+       if ((header->rcode == SERVFAIL || header->rcode == REFUSED) && forward->forwardall == 0)
+	 /* for broken servers, attempt to send to another one. */
+	 {
+	   unsigned char *pheader;
+	   size_t plen;
+	   
+	   /* recreate query from reply */
+	   pheader = find_pseudoheader(header, (size_t)n, &plen, NULL);
+	   header->ancount = htons(0);
+	   header->nscount = htons(0);
+	   header->arcount = htons(0);
+	   if ((nn = resize_packet(header, (size_t)n, pheader, plen)))
+	     {
+	       forward->forwardall = 1;
+	       header->qr = 0;
+	       header->tc = 0;
+	       forward_query(daemon, -1, NULL, NULL, 0, header, nn, now, forward);
+	       return;
+	     }
+	 }   
+
        if ((forward->sentto->flags & SERV_TYPE) == 0)
 	 {
 	   if (header->rcode == SERVFAIL || header->rcode == REFUSED)
 	     server = NULL;
 	   else
-	    {
-	      /* find good server by address if possible, otherwise assume the last one we sent to */ 
-	      struct server *last_server;
-	      for (last_server = daemon->servers; last_server; last_server = last_server->next)
-		if (!(last_server->flags & (SERV_LITERAL_ADDRESS | SERV_HAS_DOMAIN | SERV_FOR_NODOTS | SERV_NO_ADDR)) &&
-		    sockaddr_isequal(&last_server->addr, &serveraddr))
-		  {
-		    server = last_server;
-		    break;
-		  }
-	    } 
+	     {
+	       struct server *last_server;
+	       /* find good server by address if possible, otherwise assume the last one we sent to */ 
+	       for (last_server = daemon->servers; last_server; last_server = last_server->next)
+		 if (!(last_server->flags & (SERV_LITERAL_ADDRESS | SERV_HAS_DOMAIN | SERV_FOR_NODOTS | SERV_NO_ADDR)) &&
+		     sockaddr_isequal(&last_server->addr, &serveraddr))
+		   {
+		     server = last_server;
+		     break;
+		   }
+	     } 
 	   daemon->last_server = server;
 	 }
-       
-      if ((n = process_reply(daemon, header, now, forward->crc, server, (unsigned int)n)))
-	{
-	  header->id = htons(forward->orig_id);
-	  header->ra = 1; /* recursion if available */
-	  send_from(forward->fd, daemon->options & OPT_NOWILD, daemon->packet, n, 
-		    &forward->source, &forward->dest, forward->iface);
-	}
-	 
+	         
       /* If the answer is an error, keep the forward record in place in case
 	 we get a good reply from another server. Kill it when we've
 	 had replies from all to avoid filling the forwarding table when
 	 everything is broken */
       if (forward->forwardall == 0 || --forward->forwardall == 1 || 
 	  (header->rcode != REFUSED && header->rcode != SERVFAIL))
-	forward->new_id = 0; /* cancel */
+	{
+	  if ((nn = process_reply(daemon, header, now, forward->crc, server, (size_t)n)))
+	    {
+	      header->id = htons(forward->orig_id);
+	      header->ra = 1; /* recursion if available */
+	      send_from(forward->fd, daemon->options & OPT_NOWILD, daemon->packet, nn, 
+			&forward->source, &forward->dest, forward->iface);
+	    }
+	  forward->new_id = 0; /* cancel */
+	}
     }
 }
 
@@ -494,10 +520,11 @@ void receive_query(struct listener *listen, struct daemon *daemon, time_t now)
   HEADER *header = (HEADER *)daemon->packet;
   union mysockaddr source_addr;
   unsigned short type;
-  struct iname *tmp;
   struct all_addr dst_addr;
   struct in_addr netmask, dst_addr_4;
-  int m, n, if_index = 0;
+  size_t m;
+  ssize_t n;
+  int if_index = 0;
   struct iovec iov[1];
   struct msghdr msg;
   struct cmsghdr *cmptr;
@@ -506,7 +533,7 @@ void receive_query(struct listener *listen, struct daemon *daemon, time_t now)
 #ifdef HAVE_IPV6
     char control6[CMSG_SPACE(sizeof(struct in6_pktinfo))];
 #endif
-#if defined(IP_PKTINFO)
+#if defined(HAVE_LINUX_NETWORK)
     char control[CMSG_SPACE(sizeof(struct in_pktinfo))];
 #elif defined(IP_RECVDSTADDR)
     char control[CMSG_SPACE(sizeof(struct in_addr)) +
@@ -514,13 +541,19 @@ void receive_query(struct listener *listen, struct daemon *daemon, time_t now)
 #endif
   } control_u;
   
+  /* packet buffer overwritten */
+  daemon->srv_save = NULL;
+  
   if (listen->family == AF_INET && (daemon->options & OPT_NOWILD))
     {
       dst_addr_4 = listen->iface->addr.in.sin_addr;
       netmask = listen->iface->netmask;
     }
   else
-    dst_addr_4.s_addr = 0;
+    {
+      dst_addr_4.s_addr = 0;
+      netmask.s_addr = 0;
+    }
 
   iov[0].iov_base = daemon->packet;
   iov[0].iov_len = daemon->edns_pktsz;
@@ -536,13 +569,15 @@ void receive_query(struct listener *listen, struct daemon *daemon, time_t now)
   if ((n = recvmsg(listen->fd, &msg, 0)) == -1)
     return;
   
-  if (n < (int)sizeof(HEADER) || header->qr)
+  if (n < (int)sizeof(HEADER) || 
+      (msg.msg_flags & MSG_TRUNC) ||
+      header->qr)
     return;
   
   source_addr.sa.sa_family = listen->family;
 #ifdef HAVE_IPV6
   if (listen->family == AF_INET6)
-    source_addr.in6.sin6_flowinfo = htonl(0);
+    source_addr.in6.sin6_flowinfo = 0;
 #endif
   
   if (!(daemon->options & OPT_NOWILD))
@@ -552,7 +587,7 @@ void receive_query(struct listener *listen, struct daemon *daemon, time_t now)
       if (msg.msg_controllen < sizeof(struct cmsghdr))
 	return;
 
-#if defined(IP_PKTINFO)
+#if defined(HAVE_LINUX_NETWORK)
       if (listen->family == AF_INET)
 	for (cmptr = CMSG_FIRSTHDR(&msg); cmptr; cmptr = CMSG_NXTHDR(&msg, cmptr))
 	  if (cmptr->cmsg_level == SOL_IP && cmptr->cmsg_type == IP_PKTINFO)
@@ -607,36 +642,11 @@ void receive_query(struct listener *listen, struct daemon *daemon, time_t now)
 	  netmask = ((struct sockaddr_in *) &ifr.ifr_addr)->sin_addr;
 	}
 
-      for (tmp = daemon->if_except; tmp; tmp = tmp->next)
-	if (tmp->name && (strcmp(tmp->name, ifr.ifr_name) == 0))
-	  return;
-      
-      if (daemon->if_names || daemon->if_addrs)
-	{
-	  for (tmp = daemon->if_names; tmp; tmp = tmp->next)
-	    if (tmp->name && (strcmp(tmp->name, ifr.ifr_name) == 0))
-	      break;
-	  if (!tmp)
-	    for (tmp = daemon->if_addrs; tmp; tmp = tmp->next)
-	      if (tmp->addr.sa.sa_family == listen->family)
-		{
-		  if (tmp->addr.sa.sa_family == AF_INET &&
-		      tmp->addr.in.sin_addr.s_addr == dst_addr.addr.addr4.s_addr)
-		    break;
-#ifdef HAVE_IPV6
-		  else if (tmp->addr.sa.sa_family == AF_INET6 &&
-			   memcmp(&tmp->addr.in6.sin6_addr, 
-				  &dst_addr.addr.addr6, 
-				  sizeof(struct in6_addr)) == 0)
-		    break;
-#endif
-		}
-	  if (!tmp)
-	    return; 
-	}
+      if (!iface_check(daemon, listen->family, &dst_addr, ifr.ifr_name))
+	return;
     }
   
-  if (extract_request(header, (unsigned int)n, daemon->namebuff, &type))
+  if (extract_request(header, (size_t)n, daemon->namebuff, &type))
     {
       if (listen->family == AF_INET) 
 	log_query(F_QUERY | F_IPV4 | F_FORWARD, daemon->namebuff, 
@@ -648,18 +658,18 @@ void receive_query(struct listener *listen, struct daemon *daemon, time_t now)
 #endif
     }
 
-  m = answer_request (header, ((char *) header) + PACKETSZ, (unsigned int)n, daemon, 
+  m = answer_request (header, ((char *) header) + PACKETSZ, (size_t)n, daemon, 
 		      dst_addr_4, netmask, now);
   if (m >= 1)
     send_from(listen->fd, daemon->options & OPT_NOWILD, (char *)header, m, &source_addr, &dst_addr, if_index);
   else
     forward_query(daemon, listen->fd, &source_addr, &dst_addr, if_index,
-		  header, n, now);
+		  header, (size_t)n, now, NULL);
 }
 
-static int read_write(int fd, char *packet, int size, int rw)
+static int read_write(int fd, unsigned char *packet, int size, int rw)
 {
-  int n, done;
+  ssize_t n, done;
   
   for (done = 0; done < size; done += n)
     {
@@ -673,7 +683,7 @@ static int read_write(int fd, char *packet, int size, int rw)
 	return 0;
       else if (n == -1)
 	{
-	  if (retry_send())
+	  if (errno == EINTR)
 	    goto retry;
 	  else
 	    return 0;
@@ -686,14 +696,15 @@ static int read_write(int fd, char *packet, int size, int rw)
    blocking as neccessary, and then return. Note, need to be a bit careful
    about resources for debug mode, when the fork is suppressed: that's
    done by the caller. */
-char *tcp_request(struct daemon *daemon, int confd, time_t now,
-		  struct in_addr local_addr, struct in_addr netmask)
+unsigned char *tcp_request(struct daemon *daemon, int confd, time_t now,
+			   struct in_addr local_addr, struct in_addr netmask)
 {
-  int size = 0, m;
+  int size = 0;
+  size_t m;
   unsigned short qtype, gotname;
   unsigned char c1, c2;
   /* Max TCP packet + slop */
-  char *packet = malloc(65536 + MAXDNAME + RRFIXEDSZ);
+  unsigned char *packet = malloc(65536 + MAXDNAME + RRFIXEDSZ);
   HEADER *header;
   struct server *last_server;
   
@@ -874,7 +885,7 @@ static struct frec *get_new_frec(time_t now)
       if (!warntime || difftime(now, warntime) > LOGRATE)
 	{
 	  warntime = now;
-	  syslog(LOG_WARNING, "forwarding table overflow: check for server loops.");
+	  syslog(LOG_WARNING, _("forwarding table overflow: check for server loops."));
 	}
       return NULL;
     }
@@ -883,6 +894,7 @@ static struct frec *get_new_frec(time_t now)
     {
       f->next = frec_list;
       f->time = now;
+      f->new_id = 0;
       frec_list = f;
     }
   return f; /* OK if malloc fails and this is NULL */
