@@ -1,4 +1,4 @@
-/* dnsmasq is Copyright (c) 2000 - 2005 Simon Kelley
+/* dnsmasq is Copyright (c) 2000 - 2006 Simon Kelley
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -10,8 +10,6 @@
    GNU General Public License for more details.
 */
 
-/* Author's email: simon@thekelleys.org.uk */
-
 #include "dnsmasq.h"
 
 struct myoption {
@@ -21,7 +19,7 @@ struct myoption {
   int val;
 };
 
-#define OPTSTRING "31yZDNLERKzowefnbvhdkqr:m:p:c:l:s:i:t:u:g:a:x:S:C:A:T:H:Q:I:B:F:G:O:M:X:V:U:j:P:J:W:Y:2:"
+#define OPTSTRING "531yZDNLERKzowefnbvhdkqr:m:p:c:l:s:i:t:u:g:a:x:S:C:A:T:H:Q:I:B:F:G:O:M:X:V:U:j:P:J:W:Y:2:4:6:"
 
 static const struct myoption opts[] = { 
   {"version", 0, 0, 'v'},
@@ -82,6 +80,9 @@ static const struct myoption opts[] = {
   {"txt-record", 1, 0, 'Y'},
   {"enable-dbus", 0, 0, '1'},
   {"bootp-dynamic", 0, 0, '3'},
+  {"dhcp-mac", 1, 0, '4'},
+  {"no-ping", 0, 0, '5'},
+  {"dhcp-script", 1, 0, '6'},
   {0, 0, 0, 0}
 };
 
@@ -111,6 +112,7 @@ static const struct optflags optmap[] = {
   { 'y', OPT_LOCALISE },
   { '1', OPT_DBUS },
   { '3', OPT_BOOTP_DYNAMIC },
+  { '5', OPT_NO_PING },
   { 'v', 0},
   { 'w', 0},
   { 0, 0 }
@@ -177,6 +179,9 @@ static const struct {
   { "-1, --enable-dbus", gettext_noop("Enable the DBus interface for setting upstream servers, etc."), NULL },
   { "-2, --no-dhcp-interface=interface", gettext_noop("Do not provide DHCP on this interface, only provide DNS."), NULL },
   { "-3, --bootp-dynamic", gettext_noop("Enable dynamic address allocation for bootp."), NULL },
+  { "-4, --dhcp-mac=<id>,<mac address>", gettext_noop("Map MAC address (with wildcards) to option set."), NULL },
+  { "-5, --no-ping", gettext_noop("Disable ICMP echo address checking in the DHCP server."), NULL },
+  { "-6, --dhcp-script=path", gettext_noop("Script to run on DHCP lease creation and destruction."), NULL },
   { NULL, NULL, NULL }
 }; 
 
@@ -298,7 +303,6 @@ struct daemon *read_opts (int argc, char **argv, char *compile_opts)
   daemon->namebuff = buff;
 
   /* Set defaults - everything else is zero or NULL */
-  daemon->min_leasetime = UINT_MAX;
   daemon->cachesize = CACHESIZ;
   daemon->port = NAMESERVER_PORT;
   daemon->default_resolv.is_default = 1;
@@ -607,6 +611,15 @@ struct daemon *read_opts (int argc, char **argv, char *compile_opts)
 	    case 'l':
 	      daemon->lease_file = safe_string_alloc(arg);
 	      break;
+
+	    case '6':
+#ifdef NO_FORK
+	      problem = _("cannot run scripts under uClinux");
+	      option = '?';
+#else
+	      daemon->lease_change_command = safe_string_alloc(arg);
+#endif
+	      break;
 	      
 	    case 'H':
 	      {
@@ -646,8 +659,6 @@ struct daemon *read_opts (int argc, char **argv, char *compile_opts)
 		   "interface=" to disable all interfaces except loop. */
 		new->name = safe_string_alloc(arg);
 		new->isloop = new->used = 0;
-		if (safe_strchr(new->name, ':'))
-		  daemon->options |= OPT_NOWILD;
 		arg = comma;
 	      } while (arg);
 	      break;
@@ -663,8 +674,6 @@ struct daemon *read_opts (int argc, char **argv, char *compile_opts)
 		  {
 		    new->next = daemon->if_except;
 		    daemon->if_except = new;
-		    if (safe_strchr(new->name, ':'))
-		      daemon->options |= OPT_NOWILD;
 		  }
 		else
 		  {
@@ -702,16 +711,17 @@ struct daemon *read_opts (int argc, char **argv, char *compile_opts)
 		  {
 		    new->addr.sa.sa_family = AF_INET;
 #ifdef HAVE_SOCKADDR_SA_LEN
-		    new->addr.in.sin_len = sizeof(struct sockaddr_in);
+		    new->addr.in.sin_len = sizeof(new->addr.in);
 #endif
 		  }
 #ifdef HAVE_IPV6
 		else if (arg && inet_pton(AF_INET6, arg, &new->addr.in6.sin6_addr) > 0)
 		  {
 		    new->addr.sa.sa_family = AF_INET6;
-		    new->addr.in6.sin6_flowinfo = htonl(0);
+		    new->addr.in6.sin6_flowinfo = 0;
+		    new->addr.in6.sin6_scope_id = 0;
 #ifdef HAVE_SOCKADDR_SA_LEN
-		    new->addr.in6.sin6_len = sizeof(struct sockaddr_in6);
+		    new->addr.in6.sin6_len = sizeof(new->addr.in6);
 #endif
 		  }
 #endif
@@ -839,9 +849,10 @@ struct daemon *read_opts (int argc, char **argv, char *compile_opts)
 			newlist->addr.in6.sin6_port = htons(serv_port);
 			newlist->source_addr.in6.sin6_port = htons(source_port);
 			newlist->addr.sa.sa_family = newlist->source_addr.sa.sa_family = AF_INET6;
-			newlist->addr.in6.sin6_flowinfo = newlist->source_addr.in6.sin6_flowinfo = htonl(0);
+			newlist->addr.in6.sin6_flowinfo = newlist->source_addr.in6.sin6_flowinfo = 0;
+			newlist->addr.in6.sin6_scope_id = newlist->source_addr.in6.sin6_scope_id = 0;
 #ifdef HAVE_SOCKADDR_SA_LEN
-			newlist->addr.in6.sin6_len = newlist->source_addr.in6.sin6_len = sizeof(struct sockaddr_in6);
+			newlist->addr.in6.sin6_len = newlist->source_addr.in6.sin6_len = sizeof(newlist->addr.in6);
 #endif
 			if (source)
 			  {
@@ -949,6 +960,7 @@ struct daemon *read_opts (int argc, char **argv, char *compile_opts)
 		new->broadcast.s_addr = 0;
 		new->router.s_addr = 0;
 		new->netid.net = NULL;
+		new->filter = NULL;
 		new->flags = 0;
 		
 		problem = _("bad dhcp-range");
@@ -959,26 +971,40 @@ struct daemon *read_opts (int argc, char **argv, char *compile_opts)
 		    break;
 		  }
 		
-		for (cp = arg; *cp; cp++)
-		  if (!(*cp == ' ' || *cp == '.' ||  (*cp >='0' && *cp <= '9')))
-		    break;
-
-		if (*cp != ',' && (comma = strchr(arg, ',')))
+		while(1)
 		  {
-		    *comma = 0;
-		    if (strstr(arg, "net:") == arg)
+		    for (cp = arg; *cp; cp++)
+		      if (!(*cp == ' ' || *cp == '.' ||  (*cp >='0' && *cp <= '9')))
+			break;
+		    
+		    if (*cp != ',' && (comma = strchr(arg, ',')))
 		      {
-			new->netid.net = safe_string_alloc(arg+4);
-			new->netid.next = NULL;
-			new->flags |= CONTEXT_FILTER;
+			*comma = 0;
+			if (strstr(arg, "net:") == arg)
+			  {
+			    struct dhcp_netid *tt = safe_malloc(sizeof (struct dhcp_netid));
+			    tt->net = safe_string_alloc(arg+4);
+			    tt->next = new->filter;
+			    new->filter = tt;
+			  }
+			else
+			  {
+			    if (new->netid.net)
+			      {
+				option = '?';
+				problem = _("only one netid tag allowed");
+			      }
+			    else
+			      new->netid.net = safe_string_alloc(arg);
+			  }
+			arg = comma + 1;
 		      }
 		    else
-		      new->netid.net = safe_string_alloc(arg);
-		    a[0] = comma + 1;
+		      {
+			a[0] = arg;
+			break;
+		      }
 		  }
-		else
-		  a[0] = arg;
-
 		
 		for (k = 1; k < 5; k++)
 		  {
@@ -987,7 +1013,7 @@ struct daemon *read_opts (int argc, char **argv, char *compile_opts)
 		    *(a[k]++) = 0;
 		  }
 		  
-		if ((k < 2) || ((new->start.s_addr = inet_addr(a[0])) == (in_addr_t)-1))
+		if (option == '?' || (k < 2) || ((new->start.s_addr = inet_addr(a[0])) == (in_addr_t)-1))
 		  option = '?';
 		else if (strcmp(a[1], "static") == 0)
 		  {
@@ -1067,9 +1093,6 @@ struct daemon *read_opts (int argc, char **argv, char *compile_opts)
 			  }
 		      }
 		  }
-				
-		if (new->lease_time < daemon->min_leasetime)
-		  daemon->min_leasetime = new->lease_time;
 		break;
 	      }
 
@@ -1110,7 +1133,7 @@ struct daemon *read_opts (int argc, char **argv, char *compile_opts)
 			      int len;
 			      arg += 3; /* dump id: */
 			      if (strchr(arg, ':'))
-				len = parse_hex(arg, (unsigned char *)arg, -1, NULL);
+				len = parse_hex(arg, (unsigned char *)arg, -1, NULL, NULL);
 			      else
 				len = (int) strlen(arg);
 			      
@@ -1125,10 +1148,11 @@ struct daemon *read_opts (int argc, char **argv, char *compile_opts)
 			  new->flags |= CONFIG_NETID;
 			  new->netid.net = safe_string_alloc(arg+4);
 			}
-		      else if (parse_hex(a[j],  new->hwaddr, 6, &new->wildcard_mask) == 6)
+		      else 
+			{
+			  new->hwaddr_len = parse_hex(a[j],  new->hwaddr, DHCP_CHADDR_MAX, &new->wildcard_mask, &new->hwaddr_type);
 			  new->flags |= CONFIG_HWADDR;
-		      else
-			option = '?';
+			}
 		    }
 		  else if (strchr(a[j], '.') && (in.s_addr = inet_addr(a[j])) != (in_addr_t)-1)
 		    {
@@ -1208,11 +1232,8 @@ struct daemon *read_opts (int argc, char **argv, char *compile_opts)
 		    free(new);
 		  }
 		else
-		  {
-		    if ((new->flags & CONFIG_TIME) && new->lease_time < daemon->min_leasetime)
-		      daemon->min_leasetime = new->lease_time;
-		    daemon->dhcp_conf = new;
-		  }
+		  daemon->dhcp_conf = new;
+		  
 		break;
 	      }
 	      
@@ -1223,7 +1244,7 @@ struct daemon *read_opts (int argc, char **argv, char *compile_opts)
 		int addrs, digs, is_addr, is_hex, is_dec;
 		
 		new->len = 0;
-		new->is_addr = 0;
+		new->flags = 0;
 		new->netid = NULL;
 		new->val = NULL;
 		new->vendor_class = NULL;
@@ -1335,7 +1356,7 @@ struct daemon *read_opts (int argc, char **argv, char *compile_opts)
 			  digs++;
 			  is_dec = is_addr = 0;
 			}
-		      else if (*cp == '.')
+		      else if (*cp == '.' || *cp == '/')
 			is_dec = is_hex = 0;
 		      else if (!((*cp >='0' && *cp <= '9') || *cp == '-'))
 			{
@@ -1357,7 +1378,7 @@ struct daemon *read_opts (int argc, char **argv, char *compile_opts)
 		      {
 			new->len = digs;
 			new->val = safe_malloc(new->len);
-			parse_hex(comma, new->val, digs, NULL);
+			parse_hex(comma, new->val, digs, NULL, NULL);
 		      }
 		    else if (is_dec)
 		      {
@@ -1396,19 +1417,42 @@ struct daemon *read_opts (int argc, char **argv, char *compile_opts)
 		    else if (is_addr)	
 		      {
 			struct in_addr in;
-			unsigned char *op;
-			new->len = INADDRSZ * addrs;
-			new->val = op = safe_malloc(new->len);
-			new->is_addr = 1;
+       			unsigned char *op;
+			char *slash;
+			/* max length of address/subnet descriptor is five bytes */
+			new->val = op = safe_malloc(5 * addrs);
+			if (!new->vendor_class)
+			  new->flags |= DHOPT_ADDR;
 			while (addrs--) 
 			  {
 			    cp = comma;
 			    if ((comma = strchr(cp, ',')))
 			      *comma++ = 0;
+			    if ((slash = strchr(cp, '/')))
+			      *slash++ = 0;
 			    in.s_addr = inet_addr(cp);
-			    memcpy(op, &in, INADDRSZ);
-			    op += INADDRSZ;
+			    if (!slash)
+			      {
+				memcpy(op, &in, INADDRSZ);
+				op += INADDRSZ;
+			      }
+			    else
+			      {
+				unsigned char *p = (unsigned char *)&in;
+				int netsize = atoi(slash);
+				*op++ = netsize;
+				if (netsize > 0)
+				  *op++ = *p++;
+				if (netsize > 8)
+				  *op++ = *p++;
+				if (netsize > 16)
+				  *op++ = *p++;
+				if (netsize > 24)
+				 *op++ = *p++;
+				new->flags &= ~DHOPT_ADDR; /* cannot re-write descriptor format */
+			      } 
 			  }
+			new->len = op - new->val;
 		      }
 		    else
 		      {
@@ -1416,6 +1460,7 @@ struct daemon *read_opts (int argc, char **argv, char *compile_opts)
 			new->len = strlen(comma);
 			/* keep terminating zero on string */
 			new->val = (unsigned char *)safe_string_alloc(comma);
+			new->flags |= DHOPT_STRING;
 		      }
 		  }
 
@@ -1508,7 +1553,24 @@ struct daemon *read_opts (int argc, char **argv, char *compile_opts)
 		  }
 		break;
 	      }
-
+	      
+	    case '4':
+	      {
+		if (!(comma = safe_strchr(arg, ',')))
+		  option = '?';
+		else
+		  {
+		    struct dhcp_mac *new = safe_malloc(sizeof(struct dhcp_mac));
+		    *comma = 0;
+		    new->netid.net = safe_string_alloc(arg);
+		    unhide_metas(comma+1);
+		    new->hwaddr_len = parse_hex(comma+1, new->hwaddr, DHCP_CHADDR_MAX, &new->mask, &new->hwaddr_type);
+		    new->next = daemon->dhcp_macs;
+		    daemon->dhcp_macs = new;
+		  }
+	      }
+	      break;
+	      
 	    case 'U':
 	    case 'j':
 	      {
@@ -1749,9 +1811,9 @@ struct daemon *read_opts (int argc, char **argv, char *compile_opts)
 	  else
 	    die(_("bad command line options: %s."),
 #ifdef HAVE_GETOPT_LONG
-		problem ? problem : "try --help"
+		problem ? problem : _("try --help")
 #else
-		problem ? problem : "try -w"
+		problem ? problem : _("try -w")
 #endif
 		);
 	}
