@@ -1,4 +1,4 @@
-/* dnsmasq is Copyright (c) 2000-2005 Simon Kelley
+/* dnsmasq is Copyright (c) 2000-2006 Simon Kelley
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -10,8 +10,6 @@
    GNU General Public License for more details.
 */
 
-/* Author's email: simon@thekelleys.org.uk */
-
 #include "dnsmasq.h"
 
 #define BOOTREQUEST              1
@@ -44,9 +42,12 @@
 #define OPTION_CLIENT_ID         61
 #define OPTION_USER_CLASS        77
 #define OPTION_CLIENT_FQDN       81
+#define OPTION_AGENT_ID          82
 #define OPTION_SUBNET_SELECT     118
 #define OPTION_END               255
 
+#define SUBOPT_SUBNET_SELECT     5     /* RFC 3527 */
+
 #define DHCPDISCOVER             1
 #define DHCPOFFER                2
 #define DHCPREQUEST              3
@@ -57,18 +58,24 @@
 #define DHCPINFORM               8
 
 #define have_config(config, mask) ((config) && ((config)->flags & (mask))) 
+#define option_len(opt) ((int)(((unsigned char *)(opt))[1]))
+#define option_ptr(opt) ((void *)&(((unsigned char *)(opt))[2]))
 
+static unsigned int calc_time(struct dhcp_context *context, struct dhcp_config *config, 
+			      struct dhcp_lease *lease, unsigned char *opt, time_t now);
 static unsigned char *option_put(unsigned char *p, unsigned char *end, int opt, int len, unsigned int val);
-static unsigned char *option_end(unsigned char *p, unsigned char *end, struct dhcp_packet *start);
-static unsigned char *option_put_string(unsigned char *p, unsigned char *end, int opt, char *string);
+static unsigned char *option_end(unsigned char *p, unsigned char *end, 
+				 unsigned char *agent_id, struct dhcp_packet *start);
+static unsigned char *option_put_string(unsigned char *p, unsigned char *end, 
+					int opt, char *string, int null_term);
 static void bootp_option_put(struct dhcp_packet *mess, 
 			     struct dhcp_boot *boot_opts, struct dhcp_netid *netids);
-static int option_len(unsigned char *opt);
-static void *option_ptr(unsigned char *opt);
 static struct in_addr option_addr(unsigned char *opt);
 static unsigned int option_uint(unsigned char *opt, int size);
-static void log_packet(char *type, struct in_addr *addr, unsigned char *hwaddr, char *interface, char *string);
-static unsigned char *option_find(struct dhcp_packet *mess, int size, int opt_type, int minsize);
+static void log_packet(struct daemon *daemon, char *type, struct in_addr *addr, 
+		       struct dhcp_packet *mess, char *interface, char *string);
+static unsigned char *option_find(struct dhcp_packet *mess, size_t size, int opt_type, int minsize);
+static unsigned char *option_find1(unsigned char *p, unsigned char *end, int opt, int minsize);
 static unsigned char *do_req_options(struct dhcp_context *context,
 				     unsigned char *p, unsigned char *end, 
 				     unsigned char *req_options, 
@@ -76,36 +83,41 @@ static unsigned char *do_req_options(struct dhcp_context *context,
 				     char *hostname,
 				     struct dhcp_netid *netid,
 				     struct in_addr subnet_addr,
-				     unsigned char fqdn_flags);
+				     unsigned char fqdn_flags,
+				     int null_term);
 
-int dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *iface_name, 
-	       unsigned int sz, time_t now, int unicast_dest)
+
+size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *iface_name, 
+	       size_t sz, time_t now, int unicast_dest)
 {
   unsigned char *opt, *clid = NULL;
   struct dhcp_lease *ltmp, *lease = NULL;
   struct dhcp_vendor *vendor;
+  struct dhcp_mac *mac;
   struct dhcp_netid_list *id_list;
   int clid_len = 0, ignore = 0;
-  struct dhcp_packet *mess = &daemon->dhcp_packet->data;
-  unsigned char *p = mess->options + sizeof(u32); /* skip cookie */
-  unsigned char *end = (unsigned char *)(daemon->dhcp_packet + 1);
+  struct dhcp_packet *mess = daemon->dhcp_packet.iov_base;
+  unsigned char *p, *end = (unsigned char *)(mess + 1);
   char *hostname = NULL, *offer_hostname = NULL, *client_hostname = NULL;
-  int hostname_auth = 0;
+  int hostname_auth = 0, borken_opt = 0;
   unsigned char *req_options = NULL;
   char *message = NULL;
   unsigned int time;
   struct dhcp_config *config;
   struct dhcp_netid *netid = NULL;
-  struct in_addr subnet_addr;
+  struct in_addr subnet_addr, fallback;
   unsigned short fuzz = 0;
   unsigned int mess_type = 0;
-  u8 *chaddr;
   unsigned char fqdn_flags = 0;
   subnet_addr.s_addr = 0;
+  unsigned char *agent_id = NULL;
 
-  if (mess->op != BOOTREQUEST)
+  if (mess->op != BOOTREQUEST || mess->hlen > DHCP_CHADDR_MAX)
     return 0;
-  
+   
+  if (mess->htype == 0 && mess->hlen != 0)
+    return 0;
+
   /* check for DHCP rather than BOOTP */
   if ((opt = option_find(mess, sz, OPTION_MESSAGE_TYPE, 1)))
     {
@@ -115,13 +127,52 @@ int dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *iface_
       if (*((u32 *)&mess->options) != htonl(DHCP_COOKIE))
 	return 0;
 
+      /* two things to note here: expand_buf may move the packet,
+	 so reassign mess from daemon->packet. Also, the size
+	 sent includes the IP and UDP headers, hence the magic "-28" */
+      if ((opt = option_find(mess, sz, OPTION_MAXMESSAGE, 2)))
+	{
+	  size_t size = (size_t)option_uint(opt, 2) - 28;
+	  
+	  if (size > DHCP_PACKET_MAX)
+	    size = DHCP_PACKET_MAX;
+	  else if (size < sizeof(struct dhcp_packet))
+	    size = sizeof(struct dhcp_packet);
+	  
+	  if (expand_buf(&daemon->dhcp_packet, size))
+	    {
+	      mess = daemon->dhcp_packet.iov_base;
+	      end = ((unsigned char *)mess) + size;
+	    }
+	}
+
       /* Some buggy clients set ciaddr when they shouldn't, so clear that here since
 	 it can affect the context-determination code. */
       if ((option_find(mess, sz, OPTION_REQUESTED_IP, INADDRSZ) || mess_type == DHCPDISCOVER))
 	mess->ciaddr.s_addr = 0;
 
-      /* Check for RFC3011 subnet selector */
-      if ((opt = option_find(mess, sz, OPTION_SUBNET_SELECT, INADDRSZ)))
+      if ((opt = option_find(mess, sz, OPTION_AGENT_ID, 1)))
+	{
+	  /* Any agent-id needs to be copied back out, verbatim, as the last option
+	     in the packet. Here, we shift it to the very end of the buffer, if it doesn't
+	     get overwritten, then it will be shuffled back at the end of processing.
+	     Note that the incoming options must not be overwritten here, so there has to 
+	     be enough free space at the end of the packet to copy the option. */
+	  unsigned int total = option_len(opt) + 2;
+	  unsigned char *last_opt = option_find(mess, sz, OPTION_END, 0);
+	  if (last_opt && last_opt < end - total)
+	    {
+	      agent_id = end - total;
+	      memcpy(agent_id, opt, total);
+	    }
+
+	  /* look for RFC3527 Link selection sub-option */
+	  if ((opt = option_find1(option_ptr(opt), option_ptr(opt) + option_len(opt), SUBOPT_SUBNET_SELECT, INADDRSZ)))
+	    subnet_addr = option_addr(opt);
+	}
+      
+      /* Check for RFC3011 subnet selector - only if RFC3527 one not present */
+      if (subnet_addr.s_addr == 0 && (opt = option_find(mess, sz, OPTION_SUBNET_SELECT, INADDRSZ)))
 	subnet_addr = option_addr(opt);
       
       /* If there is no client identifier option, use the hardware address */
@@ -132,8 +183,7 @@ int dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *iface_
 	}
       
       /* do we have a lease in store? */
-      if (mess->htype != 0)
-	lease = lease_find_by_client(mess->chaddr, clid, clid_len);
+      lease = lease_find_by_client(mess->chaddr, mess->hlen, mess->htype, clid, clid_len);
 
       /* If this request is missing a clid, but we've seen one before, 
 	 use it again for option matching etc. */
@@ -144,38 +194,14 @@ int dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *iface_
 	}
     }
   
-  /* htype == 0 is only allowed in DHCPINFORM, this seems to be a
-     microsoftism. chaddr == NULL in that case */
-
-  if (mess->htype == 0 && mess_type == DHCPINFORM)
-    {
-      chaddr = NULL;
-      if (mess->hlen != 0)
-	return 0;
-    }
-  else
-    {
-      chaddr = mess->chaddr;
-
-      /* Token ring is supported when we have packet sockets
-	 to make the HW headers for us. We don't have the code to build
-	 token ring headers when using BPF. We rely on the fact that
-	 token ring hwaddrs are the same size as ethernet hwaddrs. */
-      
-#ifdef HAVE_BPF
-      if (mess->htype != ARPHRD_ETHER)	
-#else
-      if (mess->htype != ARPHRD_ETHER && mess->htype != ARPHRD_IEEE802)
-#endif
-	{
-	  syslog(LOG_WARNING, _("DHCP request for unsupported hardware type (%d) recieved on %s"), 
-		 mess->htype, iface_name);
-	  return 0;
-	}	
-      
-      if (mess->hlen != ETHER_ADDR_LEN)
-	return 0;
-    }
+  for (mac = daemon->dhcp_macs; mac; mac = mac->next)
+    if (mac->hwaddr_len == mess->hlen &&
+	(mac->hwaddr_type == mess->htype || mac->hwaddr_type == 0) &&
+	memcmp_masked(mac->hwaddr, mess->chaddr, mess->hlen, mac->mask))
+      {
+	mac->netid.next = netid;
+	netid = &mac->netid;
+      }
   
   /* Determine network for this packet. Our caller will have already linked all the 
      contexts which match the addresses of the receiving interface but if the 
@@ -223,9 +249,14 @@ int dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *iface_
       return 0;
     }
   
+  /* keep _a_ local address available. */
+  fallback = context->local;
+  
   mess->op = BOOTREPLY;
-    
-  config = find_config(daemon->dhcp_conf, context, clid, clid_len, chaddr, NULL);
+  p = mess->options + sizeof(u32); /* skip cookie */
+  
+  config = find_config(daemon->dhcp_conf, context, clid, clid_len, 
+		       mess->chaddr, mess->hlen, mess->htype, NULL);
   
   if (mess_type == 0)
     {
@@ -233,6 +264,10 @@ int dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *iface_
       struct dhcp_netid id;
       char save = mess->file[128];
       struct in_addr *logaddr = NULL;
+
+      /* must have a MAC addr for bootp */
+      if (mess->htype == 0 || mess->hlen == 0)
+	return 0;
       
       if (have_config(config, CONFIG_DISABLE))
 	message = _("disabled");
@@ -258,7 +293,7 @@ int dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *iface_
 	}
       
       for (id_list = daemon->dhcp_ignore; id_list; id_list = id_list->next)
-	if (match_netid(id_list->list, netid))
+	if (match_netid(id_list->list, netid, 0))
 	  message = _("disabled");
       
       if (!message)
@@ -268,14 +303,16 @@ int dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *iface_
 	      logaddr = &config->addr;
 	      mess->yiaddr = config->addr;
 	      if ((lease = lease_find_by_addr(config->addr)) &&
-		  memcmp(lease->hwaddr, chaddr, ETHER_ADDR_LEN) != 0)
+		  (lease->hwaddr_len != mess->hlen ||
+		   lease->hwaddr_type != mess->htype ||
+		   memcmp(lease->hwaddr, mess->chaddr, lease->hwaddr_len) != 0))
 		message = _("address in use");
 	    }
 	  else if (!(daemon->options & OPT_BOOTP_DYNAMIC))
 	    message = _("no address configured");
 	  else
 	    {
-	      if (!(lease = lease_find_by_client(mess->chaddr, NULL, 0)) ||
+	      if (!(lease = lease_find_by_client(mess->chaddr, mess->hlen, mess->htype, NULL, 0)) ||
 		  !address_available(context, lease->addr))
 		{
 		   if (lease)
@@ -284,14 +321,16 @@ int dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *iface_
 		       lease_prune(lease, now);
 		       lease = NULL;
 		     }
-		   if (!address_allocate(context, daemon, &mess->yiaddr, chaddr, netid, now))
+		   if (!address_allocate(context, daemon, &mess->yiaddr, mess->chaddr, mess->hlen, netid, now))
 		     message = _("no address available");
 		}
 	      else
 		mess->yiaddr = lease->addr;
 	    }
 	  
-	  if (!message && !lease && (!(lease = lease_allocate(chaddr, NULL, 0, mess->yiaddr))))
+	  if (!message && 
+	      !lease && 
+	      (!(lease = lease_allocate(mess->yiaddr))))
 	    message = _("no leases left");
 	  
 	  if (!message && !(context = narrow_context(context, mess->yiaddr)))
@@ -301,27 +340,27 @@ int dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *iface_
 	    {
 	      logaddr = &mess->yiaddr;
 		
-	      if (context->netid.net && !(context->flags & CONTEXT_FILTER))
+	      if (context->netid.net)
 		{
 		  context->netid.next = netid;
 		  netid = &context->netid;
 		}	 
 	      
-	      lease_set_hwaddr(lease, chaddr, NULL, 0);
+	      lease_set_hwaddr(lease, mess->chaddr, NULL, mess->hlen, mess->htype, 0);
 	      if (hostname)
 		lease_set_hostname(lease, hostname, daemon->domain_suffix, 1); 
-	      lease_set_expires(lease, 0); /* infinite lease */
+	      lease_set_expires(lease, 0xffffffff, now); /* infinite lease */
 	      
 	      p = do_req_options(context, p, end, NULL, daemon, 
-				 hostname, netid, subnet_addr, fqdn_flags);
+				 hostname, netid, subnet_addr, fqdn_flags, 0);
 	      /* must do this after do_req_options since it overwrites filename field. */
 	      mess->siaddr = context->local;
 	      bootp_option_put(mess, daemon->boot_config, netid);
-	      p = option_end(p, end, mess);
+	      p = option_end(p, end, NULL, mess);
 	    }
 	}
       
-      log_packet(NULL, logaddr, chaddr, iface_name, message);
+      log_packet(daemon, NULL, logaddr, mess, iface_name, message);
       mess->file[128] = save;
 
       if (message)
@@ -343,7 +382,7 @@ int dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *iface_
       pp = op;
       
       /* Always force update, since the client has no way to do it itself. */
-      if (fqdn_flags & 0x01)
+      if (!(fqdn_flags & 0x01))
 	fqdn_flags |= 0x02;
       
       fqdn_flags &= ~0x08;
@@ -360,6 +399,8 @@ int dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *iface_
       else
 	{
 	  memcpy(pq, op, len);
+	  if (len > 0 && op[len-1] == 0)
+	    borken_opt = 1;
 	  pq += len + 1;
 	}
       
@@ -375,8 +416,13 @@ int dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *iface_
     {
       int len = option_len(opt);
       memcpy(daemon->dhcp_buff, option_ptr(opt), len);
-      /* May not be zero terminated */
-      daemon->dhcp_buff[len] = 0;
+      /* Microsoft clients are broken, and need zero-terminated strings
+	 in options. We detect this state here, and do the same in
+	 any options we send */
+      if (len > 0 && daemon->dhcp_buff[len-1] == 0)
+	borken_opt = 1;
+      else
+	daemon->dhcp_buff[len] = 0;
       if (canonicalise(daemon->dhcp_buff))
 	client_hostname = daemon->dhcp_buff;
     }
@@ -395,7 +441,9 @@ int dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *iface_
       /* Search again now we have a hostname. 
 	 Only accept configs without CLID and HWADDR here, (they won't match)
 	 to avoid impersonation by name. */
-      struct dhcp_config *new = find_config(daemon->dhcp_conf, context, NULL, 0, chaddr, hostname);
+      struct dhcp_config *new = find_config(daemon->dhcp_conf, context, NULL, 0,
+					    mess->chaddr, mess->hlen, 
+					    mess->htype, hostname);
       if (!have_config(new, CONFIG_CLID) && !have_config(new, CONFIG_HWADDR))
 	config = new;
     }
@@ -439,10 +487,10 @@ int dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *iface_
 	      break;
 	    }
       }
-
+  
   /* if all the netids in the ignore list are present, ignore this client */
   for (id_list = daemon->dhcp_ignore; id_list; id_list = id_list->next)
-    if (match_netid(id_list->list, netid))
+    if (match_netid(id_list->list, netid, 0))
       ignore = 1;
   
   /* Can have setting to ignore the client ID for a particular MAC address or hostname */
@@ -482,7 +530,7 @@ int dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *iface_
       if (!(opt = option_find(mess, sz, OPTION_REQUESTED_IP, INADDRSZ)))
 	return 0;
       
-      log_packet("DECLINE", option_ptr(opt), chaddr, iface_name, message);
+      log_packet(daemon, "DECLINE", option_ptr(opt), mess, iface_name, message);
       
       if (lease && lease->addr.s_addr == option_addr(opt).s_addr)
 	lease_prune(lease, now);
@@ -510,50 +558,66 @@ int dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *iface_
       else
 	message = _("unknown lease");
 
-      log_packet("RELEASE", &mess->ciaddr, chaddr, iface_name, message);
+      log_packet(daemon, "RELEASE", &mess->ciaddr, mess, iface_name, message);
 	
       return 0;
       
     case DHCPDISCOVER:
-      {
-	struct in_addr addr;
-	
-	if ((opt = option_find(mess, sz, OPTION_REQUESTED_IP, INADDRSZ)))	 
-	  addr = option_addr(opt);
-	if (ignore || have_config(config, CONFIG_DISABLE))
+      if (ignore || have_config(config, CONFIG_DISABLE))
+	{
 	  message = _("ignored");
-	else if (have_config(config, CONFIG_ADDR) && 
-		 (!(ltmp = lease_find_by_addr(config->addr)) || ltmp == lease))
-	  mess->yiaddr = config->addr;
-	else if (lease && address_available(context, lease->addr))
-	  mess->yiaddr = lease->addr;
-	else if (opt && address_available(context, addr) && !lease_find_by_addr(addr) && 
-		 !config_find_by_address(daemon->dhcp_conf, addr))
-	  mess->yiaddr = addr;
-	else if (!address_allocate(context, daemon, &mess->yiaddr, chaddr, netid, now))
-	  message = _("no address available");      
-	log_packet("DISCOVER", opt ? &addr : NULL, chaddr, iface_name, message);          
-      }
+	  opt = NULL;
+	}
+      else 
+	{
+	  struct in_addr addr, conf;
+	  
+	  if ((opt = option_find(mess, sz, OPTION_REQUESTED_IP, INADDRSZ)))	 
+	    addr = option_addr(opt);
+	  
+	  conf.s_addr = 0;
+	  if (have_config(config, CONFIG_ADDR))
+	    {
+	      if ((ltmp = lease_find_by_addr(config->addr)) && ltmp != lease)
+		syslog(LOG_WARNING, _("not using configured address %s because it is leased to %s"),
+		       inet_ntoa(config->addr), print_mac(daemon, ltmp->hwaddr, ltmp->hwaddr_len));
+	      else
+		{
+		  struct dhcp_context *tmp;
+		  for (tmp = context; tmp; tmp = tmp->current)
+		    if (context->local.s_addr == config->addr.s_addr)
+		      break;
+		  if (tmp)
+		    syslog(LOG_WARNING, _("not using configured address %s because it is in use by the server"),
+			   inet_ntoa(config->addr));
+		  else
+		    conf = config->addr;
+		}
+	    }
+	  
+	  if (conf.s_addr)
+	    mess->yiaddr = conf;
+	  else if (lease && address_available(context, lease->addr))
+	    mess->yiaddr = lease->addr;
+	  else if (opt && address_available(context, addr) && !lease_find_by_addr(addr) && 
+		   !config_find_by_address(daemon->dhcp_conf, addr))
+	    mess->yiaddr = addr;
+	  else if (!address_allocate(context, daemon, &mess->yiaddr, mess->chaddr, mess->hlen, netid, now))
+	    message = _("no address available");      
+	}
+      
+      log_packet(daemon, "DISCOVER", opt ? (struct in_addr *)option_ptr(opt) : NULL, mess, iface_name, message); 
 
       if (message || !(context = narrow_context(context, mess->yiaddr)))
 	return 0;
 
-      if (context->netid.net && !(context->flags & CONTEXT_FILTER))
+      if (context->netid.net)
 	{
 	  context->netid.next = netid;
 	  netid = &context->netid;
 	}
        
-      time = have_config(config, CONFIG_TIME) ? config->lease_time : context->lease_time;
-      if ((opt = option_find(mess, sz, OPTION_LEASE_TIME, 4)))
-	{
-	  unsigned int req_time = option_uint(opt, 4);
-	  if (time == 0xffffffff || (req_time != 0xffffffff && req_time < time))
-	    time = req_time;
-	}
-      else if (lease && lease->expires != 0)
-	time = (unsigned int)difftime(lease->expires, now);
-
+      time = calc_time(context, config, lease, option_find(mess, sz, OPTION_LEASE_TIME, 4), now);
       mess->siaddr = context->local;
       bootp_option_put(mess, daemon->boot_config, netid);
       p = option_put(p, end, OPTION_MESSAGE_TYPE, 1, DHCPOFFER);
@@ -566,10 +630,10 @@ int dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *iface_
 	  p = option_put(p, end, OPTION_T2, 4, (time*7)/8);
 	}
       p = do_req_options(context, p, end, req_options, daemon, 
-			 offer_hostname, netid, subnet_addr, fqdn_flags);
-      p = option_end(p, end, mess);
+			 offer_hostname, netid, subnet_addr, fqdn_flags, borken_opt);
+      p = option_end(p, end, agent_id, mess);
       
-      log_packet("OFFER" , &mess->yiaddr, chaddr, iface_name, NULL);
+      log_packet(daemon, "OFFER" , &mess->yiaddr, mess, iface_name, NULL);
       return p - (unsigned char *)mess;
       
     case DHCPREQUEST:
@@ -610,8 +674,12 @@ int dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *iface_
       else
 	{
 	  /* RENEWING or REBINDING */ 
-	  /* Must exist a lease for this address */
-	  if (!lease || mess->ciaddr.s_addr != lease->addr.s_addr)
+	  /* Check existing lease for this address.
+	     We allow it to be missing if dhcp-authoritative mode
+	     as long as we can allocate the lease now - checked below.
+	     This makes for a smooth recovery from a lost lease DB */
+	  if ((lease && mess->ciaddr.s_addr != lease->addr.s_addr) ||
+	      (!lease && !(daemon->options & OPT_AUTHORITATIVE)))
 	    {
 	      message = _("lease not found");
 	      /* ensure we broadcast NAK */
@@ -622,12 +690,18 @@ int dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *iface_
 	  mess->yiaddr = mess->ciaddr;
 	}
       
-      log_packet("REQUEST", &mess->yiaddr, chaddr, iface_name, NULL);
+      log_packet(daemon, "REQUEST", &mess->yiaddr, mess, iface_name, NULL);
  
       if (!message)
 	{
 	  struct dhcp_config *addr_config;
-
+	  struct dhcp_context *tmp = NULL;
+	  
+	  if (have_config(config, CONFIG_ADDR))
+	    for (tmp = context; tmp; tmp = tmp->current)
+	      if (context->local.s_addr == config->addr.s_addr)
+		break;
+	  
 	  if (!(context = narrow_context(context, mess->yiaddr)))
 	    {
 	      /* If a machine moves networks whilst it has a lease, we catch that here. */
@@ -644,8 +718,8 @@ int dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *iface_
 	  /* Check if a new static address has been configured. Be very sure that
 	     when the client does DISCOVER, it will get the static address, otherwise
 	     an endless protocol loop will ensue. */
-	  
-	  else if (have_config(config, CONFIG_ADDR) && 
+	  else if (!tmp &&
+		   have_config(config, CONFIG_ADDR) && 
 		   config->addr.s_addr != mess->yiaddr.s_addr &&
 		   (!(ltmp = lease_find_by_addr(config->addr)) || ltmp == lease))
 	    message = _("static lease available");
@@ -657,19 +731,24 @@ int dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *iface_
 	  else if ((ltmp = lease_find_by_addr(mess->yiaddr)) && ltmp != lease)
 	    message = _("address in use");
 	  
-	  else if (!lease && !(lease = lease_allocate(chaddr, clid, clid_len, mess->yiaddr)))
+	  else if (!clid && mess->hlen == 0)
+	    message = _("no unique-id");
+	  
+	  else if (!lease && 
+		   !(lease = lease_allocate(mess->yiaddr)))
 	    message = _("no leases left");
 	}
       
       if (message)
 	{
-	  log_packet("NAK", &mess->yiaddr, chaddr, iface_name, message);
+	  log_packet(daemon, "NAK", &mess->yiaddr, mess, iface_name, message);
 	  
 	  mess->siaddr.s_addr = mess->yiaddr.s_addr = 0;
 	  bootp_option_put(mess, NULL, NULL);
 	  p = option_put(p, end, OPTION_MESSAGE_TYPE, 1, DHCPNAK);
-	  p = option_put(p, end, OPTION_SERVER_IDENTIFIER, INADDRSZ, ntohl(context->local.s_addr));
-	  p = option_put_string(p, end, OPTION_MESSAGE, message);
+	  p = option_put(p, end, OPTION_SERVER_IDENTIFIER, INADDRSZ, 
+			 ntohl(context ? context->local.s_addr : fallback.s_addr));
+	  p = option_put_string(p, end, OPTION_MESSAGE, message, borken_opt);
 	  /* This fixes a problem with the DHCP spec, broadcasting a NAK to a host on 
 	     a distant subnet which unicast a REQ to us won't work. */
 	  if (!unicast_dest || mess->giaddr.s_addr != 0 || 
@@ -687,26 +766,19 @@ int dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *iface_
 	      hostname_auth = 1;
 	    }
 
-	  log_packet("ACK", &mess->yiaddr, chaddr, iface_name, hostname);
+	  log_packet(daemon, "ACK", &mess->yiaddr, mess, iface_name, hostname);
       
-	  if (context->netid.net && !(context->flags & CONTEXT_FILTER))
+	  if (context->netid.net)
 	    {
 	      context->netid.next = netid;
 	      netid = &context->netid;
 	    }
 	
-	  time = have_config(config, CONFIG_TIME) ? config->lease_time : context->lease_time;
-	  if ((opt = option_find(mess, sz, OPTION_LEASE_TIME, 4)))
-	    {
-	      unsigned int req_time = option_uint(opt, 4);
-	      if (time == 0xffffffff || (req_time != 0xffffffff && req_time < time))
-		time = req_time;
-	    }
-	  
-	  lease_set_hwaddr(lease, chaddr, clid, clid_len);
+	  time = calc_time(context, config, NULL, option_find(mess, sz, OPTION_LEASE_TIME, 4), now);
+	  lease_set_hwaddr(lease, mess->chaddr, clid, mess->hlen, mess->htype, clid_len);
 	  if (hostname)
 	    lease_set_hostname(lease, hostname, daemon->domain_suffix, hostname_auth);
-	  lease_set_expires(lease, time == 0xffffffff ? 0 : now + (time_t)time);
+	  lease_set_expires(lease, time, now);
 	  	  
 	  mess->siaddr = context->local;
 	  bootp_option_put(mess, daemon->boot_config, netid);
@@ -718,20 +790,20 @@ int dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *iface_
 	      while (fuzz > (time/16))
 		fuzz = fuzz/2; 
 	      p = option_put(p, end, OPTION_T1, 4, (time/2) - fuzz);
-	      p = option_put(p, end, OPTION_T2, 4, ((time * 7)/8) - fuzz);
+	      p = option_put(p, end, OPTION_T2, 4, ((time/8)*7) - fuzz);
 	    }
 	  p = do_req_options(context, p, end, req_options, daemon, 
-			     hostname, netid, subnet_addr, fqdn_flags);
+			     hostname, netid, subnet_addr, fqdn_flags, borken_opt);
 	}
 
-      p = option_end(p, end, mess);
+      p = option_end(p, end, agent_id, mess);
       return p - (unsigned char *)mess; 
       
     case DHCPINFORM:
       if (ignore || have_config(config, CONFIG_DISABLE))
 	message = _("ignored");
       
-      log_packet("INFORM", &mess->ciaddr, chaddr, iface_name, message);
+      log_packet(daemon, "INFORM", &mess->ciaddr, mess, iface_name, message);
      
       if (message || mess->ciaddr.s_addr == 0 || 
 	  !(context = narrow_context(context, mess->ciaddr)))
@@ -750,44 +822,57 @@ int dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *iface_
       if (!hostname)
 	hostname = host_from_dns(daemon, mess->yiaddr);
       p = do_req_options(context, p, end, req_options, daemon, 
-			 hostname, netid, subnet_addr, fqdn_flags);
-      p = option_end(p, end, mess);
+			 hostname, netid, subnet_addr, fqdn_flags, borken_opt);
+      p = option_end(p, end, agent_id, mess);
       
-      log_packet("ACK", &mess->ciaddr, chaddr, iface_name, hostname);
+      log_packet(daemon, "ACK", &mess->ciaddr, mess, iface_name, hostname);
       return p - (unsigned char *)mess; 
     }
   
   return 0;
 }
 
-static void log_packet(char *type, struct in_addr *addr, unsigned char *hwaddr, char *interface, char *string)
+static unsigned int calc_time(struct dhcp_context *context, struct dhcp_config *config, 
+			      struct dhcp_lease *lease, unsigned char *opt, time_t now)
 {
-  u8 empty[] = { 0, 0, 0, 0, 0, 0};
+  unsigned int time = have_config(config, CONFIG_TIME) ? config->lease_time : context->lease_time;
+  
+  if (opt)
+    { 
+      unsigned int req_time = option_uint(opt, 4);
+      if (req_time < 120 )
+	req_time = 120; /* sanity */
+      if (time == 0xffffffff || (req_time != 0xffffffff && req_time < time))
+	time = req_time;
+    }
+  else if (lease && lease->expires != 0 && difftime(lease->expires, now) > 0.0)
+    {
+      unsigned int lease_time = (unsigned int)difftime(lease->expires, now);
+      
+      /* put a floor on lease-remaining time. */
+      if (lease_time < 360 )
+	lease_time = 360;
+      
+      if (time > lease_time)
+	time = lease_time;
+    }
 
-  if (!hwaddr)
-    hwaddr = empty;
+  return time;
+}
 
-  syslog(LOG_INFO, "%s%s(%s)%s%s %.2x:%.2x:%.2x:%.2x:%.2x:%.2x%s%s",
+static void log_packet(struct daemon *daemon, char *type, struct in_addr *addr, 
+		       struct dhcp_packet *mess, char *interface, char *string)
+{
+  syslog(LOG_INFO, "%s%s(%s) %s%s%s %s",
 	 type ? "DHCP" : "BOOTP",
 	 type ? type : "",
 	 interface, 
-	 addr ? " " : "",
 	 addr ? inet_ntoa(*addr) : "",
-	 hwaddr[0], hwaddr[1], hwaddr[2], hwaddr[3], hwaddr[4], hwaddr[5],
-	 string ? " " : "",
+	 addr ? " " : "",
+	 print_mac(daemon, mess->chaddr, mess->hlen),
 	 string ? string : "");
 }
 
-static int option_len(unsigned char *opt)
-{
-  return opt[1];
-}
-
-static void *option_ptr(unsigned char *opt)
-{
-  return &opt[2];
-}
-
 static struct in_addr option_addr(unsigned char *opt)
 {
   /* this worries about unaligned data in the option. */
@@ -818,12 +903,12 @@ static void bootp_option_put(struct dhcp_packet *mess,
   struct dhcp_boot *tmp;
   
   for (tmp = boot_opts; tmp; tmp = tmp->next)
-    if (match_netid(tmp->netid, netids))
+    if (match_netid(tmp->netid, netids, 0))
       break;
   if (!tmp)
     /* No match, look for one without a netid */
     for (tmp = boot_opts; tmp; tmp = tmp->next)
-      if (!tmp->netid)
+      if (match_netid(tmp->netid, netids, 1))
 	break;
 
   /* Do this _after_ the matching above, since in 
@@ -871,8 +956,16 @@ static unsigned char *option_put(unsigned char *p, unsigned char *end, int opt,
   return p;
 }
 
-static unsigned char *option_end(unsigned char *p, unsigned char *end, struct dhcp_packet *start)
+static unsigned char *option_end(unsigned char *p, unsigned char *end, 
+				 unsigned char *agent_id, struct dhcp_packet *start)
 {
+  /* shuffle agent-id back down if we have room for it */
+  if (agent_id && p < agent_id)
+    {
+      memmove(p, agent_id, end - agent_id);
+      p += end - agent_id;
+    }
+
   *(p++) = OPTION_END;
   while ((p < end) && (p - ((unsigned char *)start) < MIN_PACKETSZ))
     *p++ = 0;
@@ -880,10 +973,14 @@ static unsigned char *option_end(unsigned char *p, unsigned char *end, struct dh
   return p;
 }
 
-static unsigned char *option_put_string(unsigned char *p, unsigned char *end, int opt, char *string)
+static unsigned char *option_put_string(unsigned char *p, unsigned char *end, int opt, 
+					char *string, int null_term)
 {
   size_t len = strlen(string);
 
+  if (null_term && len != 255)
+    len++;
+
   if (check_space(p, end, len, opt))
     {
       *(p++) = opt;
@@ -895,61 +992,54 @@ static unsigned char *option_put_string(unsigned char *p, unsigned char *end, in
   return p;
 }
  
-static unsigned char *option_find1(unsigned char *p, unsigned char *end, int opt, int *overload)
+static unsigned char *option_find1(unsigned char *p, unsigned char *end, int opt, int minsize)
 {
-  if (!p)
-    return NULL;
-  
   while (*p != OPTION_END) 
     {
       if (p >= end)
-	return 0; /* malformed packet */
+	return NULL; /* malformed packet */
       else if (*p == OPTION_PAD)
 	p++;
-      else if (*p == OPTION_OVERLOAD)
-	{
-	  if (p >= end - 3)
-	    return 0; /* malformed packet */
-	  if (overload) 
-	    *overload = *(p+2);
-	  p += 3;
-	}
       else 
 	{ 
-	  int opt_len;;
+	  int opt_len;
 	  if (p >= end - 2)
-	    return 0; /* malformed packet */
+	    return NULL; /* malformed packet */
 	  opt_len = option_len(p);
 	  if (p >= end - (2 + opt_len))
-	    return 0; /* malformed packet */
-	  if (*p == opt)
+	    return NULL; /* malformed packet */
+	  if (*p == opt && opt_len >= minsize)
 	    return p;
 	  p += opt_len + 2;
 	}
     }
   
-  return NULL;
+  return opt == OPTION_END ? p : NULL;
 }
  
-static unsigned char *option_find(struct dhcp_packet *mess, int size, int opt_type, int minsize)
+static unsigned char *option_find(struct dhcp_packet *mess, size_t size, int opt_type, int minsize)
 {
-  int overload = 0; 
-  unsigned char *ret;
+  unsigned char *ret, *opt;
   
   /* skip over DHCP cookie; */
-  ret = option_find1(&mess->options[0] + sizeof(u32), ((unsigned char *)mess) + size, opt_type, &overload);
-  
-  if (!ret && (overload & 1)) 
-    ret = option_find1(&mess->file[0], &mess->file[128], opt_type, &overload);
+  if ((ret = option_find1(&mess->options[0] + sizeof(u32), ((unsigned char *)mess) + size, opt_type, minsize)))
+    return ret;
 
-  if (!ret && (overload & 2))
-    ret = option_find1(&mess->sname[0], &mess->file[64], opt_type, &overload);
+  /* look for overload option. */
+  if (!(opt = option_find1(&mess->options[0] + sizeof(u32), ((unsigned char *)mess) + size, OPTION_OVERLOAD, 1)))
+    return NULL;
   
-  /* Check the option field is big enough */
-  if (ret && (option_len(ret) < minsize))
-    ret = NULL;
-  
-  return ret;
+  /* Can we look in filename area ? */
+  if ((option_uint(opt, 1) & 1) &&
+      (ret = option_find1(&mess->file[0], &mess->file[128], opt_type, minsize)))
+    return ret;
+
+  /* finally try sname area */
+  if ((option_uint(opt, 1) & 2) &&
+      (ret = option_find1(&mess->sname[0], &mess->sname[64], opt_type, minsize)))
+    return ret;
+
+  return NULL;
 }
 
 static int in_list(unsigned char *list, int opt)
@@ -972,31 +1062,30 @@ static struct dhcp_opt *option_find2(struct dhcp_netid *netid, struct dhcp_opt *
   struct dhcp_opt *tmp;  
   for (tmp = opts; tmp; tmp = tmp->next)
     if (tmp->opt == opt)
-      {
-	if (netid)
-	  {
-	    if (match_netid(tmp->netid, netid))
-	      return tmp;
-	  }
-	else if (!tmp->netid)
-	  return tmp;
-      }
+      if (match_netid(tmp->netid, netid, 1) || match_netid(tmp->netid, netid, 0))
+	return tmp;
 	      
   return netid ? option_find2(NULL, opts, opt) : NULL;
 }
 
-static unsigned char *do_opt(struct dhcp_opt *opt, unsigned char *p, unsigned char *end,  struct in_addr local)
+static unsigned char *do_opt(struct dhcp_opt *opt, unsigned char *p, unsigned char *end,  
+			     struct in_addr local, int null_term)
 {
-  if (!check_space(p, end, opt->len, opt->opt))
+  int len = opt->len;
+
+  if ((opt->flags & DHOPT_STRING) && null_term && len != 255)
+    len++;
+
+  if (!check_space(p, end, len, opt->opt))
     return p;
     
   *(p++) = opt->opt;
-  *(p++) = opt->len;
+  *(p++) = len;
   
-  if (opt->len == 0)
+  if (len == 0)
     return p;
       
-  if (opt->is_addr && !opt->vendor_class)
+  if (opt->flags & DHOPT_ADDR)
     {
       int j;
       struct in_addr *a = (struct in_addr *)opt->val;
@@ -1012,8 +1101,8 @@ static unsigned char *do_opt(struct dhcp_opt *opt, unsigned char *p, unsigned ch
     }
   else
     {
-      memcpy(p, opt->val, opt->len);
-      p += opt->len;
+      memcpy(p, opt->val, len);
+      p += len;
     }
 
   return p;
@@ -1026,13 +1115,11 @@ static unsigned char *do_req_options(struct dhcp_context *context,
 				     char *hostname,
 				     struct dhcp_netid *netid,
 				     struct in_addr subnet_addr,
-				     unsigned char fqdn_flags)
+				     unsigned char fqdn_flags,
+				     int null_term)
 {
   struct dhcp_opt *opt, *config_opts = daemon->dhcp_opts;
   char *vendor_class = NULL;
-
-  if (in_list(req_options, OPTION_MAXMESSAGE))
-    p = option_put(p, end, OPTION_MAXMESSAGE, 2, end - (unsigned char *)daemon->dhcp_packet);
   
   /* rfc3011 says this doesn't need to be in the requested options list. */
   if (subnet_addr.s_addr)
@@ -1062,7 +1149,7 @@ static unsigned char *do_req_options(struct dhcp_context *context,
   
   if (daemon->domain_suffix && in_list(req_options, OPTION_DOMAINNAME) && 
       !option_find2(netid, config_opts, OPTION_DOMAINNAME))
-    p = option_put_string(p, end, OPTION_DOMAINNAME, daemon->domain_suffix);
+    p = option_put_string(p, end, OPTION_DOMAINNAME, daemon->domain_suffix, null_term);
  
   /* Note that we ignore attempts to set the hostname using 
      --dhcp-option=12,<name> and the fqdn using
@@ -1070,14 +1157,16 @@ static unsigned char *do_req_options(struct dhcp_context *context,
   if (hostname)
     {
       if (in_list(req_options, OPTION_HOSTNAME))
-	p = option_put_string(p, end, OPTION_HOSTNAME, hostname);
+	p = option_put_string(p, end, OPTION_HOSTNAME, hostname, null_term);
       
       if (fqdn_flags != 0)
 	{
 	  int len = strlen(hostname) + 3;
 	  if (fqdn_flags & 0x04)
 	    len += 2;
-	  
+	  else if (null_term)
+	    len++;
+
 	  if (daemon->domain_suffix)
 	    len += strlen(daemon->domain_suffix) + 1;
 	  
@@ -1105,7 +1194,9 @@ static unsigned char *do_req_options(struct dhcp_context *context,
 		      *(p++) = '.';
 		      memcpy(p, daemon->domain_suffix, strlen(daemon->domain_suffix));
 		      p += strlen(daemon->domain_suffix);
-		    } 
+		    }
+		  if (null_term)
+		    *(p++) = 0;
 		}
 	    }
 	}
@@ -1134,13 +1225,13 @@ static unsigned char *do_req_options(struct dhcp_context *context,
       if (opt->opt == OPTION_VENDOR_ID)
 	vendor_class = (char *)opt->val; 
       else
-	p = do_opt(opt, p, end, context->local);
+	p = do_opt(opt, p, end, context->local, null_term);
     }  
 
   if (in_list(req_options, OPTION_VENDOR_ID))
     {      
       for (opt = daemon->vendor_opts; opt; opt = opt->next)
-	if (!opt->netid || match_netid(opt->netid, netid))
+	if (match_netid(opt->netid, netid, 1) || match_netid(opt->netid, netid, 0))
 	  {
 	    if (vendor_class && strcmp(vendor_class, (char *)opt->vendor_class) != 0)
 	      syslog(LOG_WARNING, _("More than one vendor class matches, using %s"), vendor_class);
@@ -1150,7 +1241,7 @@ static unsigned char *do_req_options(struct dhcp_context *context,
 
       if (vendor_class)
 	{
-	  p = option_put_string(p, end, OPTION_VENDOR_ID, vendor_class);
+	  p = option_put_string(p, end, OPTION_VENDOR_ID, vendor_class, 0);
 	  
 	  if (in_list(req_options, OPTION_VENDOR_CLASS_OPT))
 	    {
@@ -1168,9 +1259,9 @@ static unsigned char *do_req_options(struct dhcp_context *context,
 	      plen = p++; /* fill in later */
 	      
 	      for (opt = daemon->vendor_opts; opt; opt = opt->next)
-		if ((!opt->netid || match_netid(opt->netid, netid)) && 
+		if ((match_netid(opt->netid, netid, 1) || match_netid(opt->netid, netid, 0)) && 
 		    strcmp(vendor_class, (char *)opt->vendor_class) == 0)
-		  p = do_opt(opt, p, oend, context->local);
+		  p = do_opt(opt, p, oend, context->local, null_term);
 	      
 	      *plen = p - plen - 1;
 	    }
